"""Informix (Informix JDBC) source class."""

from typing import Any, Literal

import jaydebeapi
import pandas as pd
from pydantic import BaseModel, SecretStr

from viadot.config import get_source_credentials
from viadot.sources.base import SQL, Record
from viadot.utils import add_viadot_metadata_columns, validate


class InformixCredentials(BaseModel):
    user: str
    password: str | SecretStr | None = None
    server: str
    port: int = 1504
    db_name: str
    informix_server: str
    db_locale: str = "en_US.57372"
    client_locale: str = "en_US.819"
    driver_class: str = "com.informix.jdbc.IfxDriver"


class Informix(SQL):
    DEFAULT_SCHEMA = "informix"

    def __init__(
        self,
        credentials: InformixCredentials | None = None,
        config_key: str | None = None,
        *args,
        **kwargs,
    ):
        """Connector for Informix Informix over JDBC."""
        raw_creds = credentials or get_source_credentials(config_key) or {}
        validated_creds = InformixCredentials(**raw_creds).dict(by_alias=True)

        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.server = self.credentials.get("server")
        self.port = self.credentials.get("port")
        self.db_name = self.credentials.get("db_name")
        self.informix_server = self.credentials.get("informix_server")
        self.db_locale = self.credentials.get("db_locale")
        self.client_locale = self.credentials.get("client_locale")
        self.driver_class = self.credentials.get("driver_class")
        self.user = self.credentials.get("user")
        self.password = self.credentials.get("password")

    @property
    def conn_str(self) -> str:
        """Generate a JDBC URL for Informix.

        Returns:
            str: The JDBC URL.
        """
        return (
            f"jdbc:informix-sqli://{self.server}:{self.port}/{self.db_name}:"
            f"INFORMIXSERVER={self.informix_server};"
            f"DB_LOCALE={self.db_locale};"
            f"CLIENT_LOCALE={self.client_locale};"
        )

    @property
    def con(self) -> jaydebeapi.Connection:
        """Create and cache a JDBC connection.

        Returns:
            jaydebeapi.Connection: The JDBC connection.
        """
        if not self._con:
            password = (
                self.password.get_secret_value()
                if isinstance(self.password, SecretStr)
                else self.password
            )
            connect_args = [
                self.driver_class,
                self.conn_str,
                [self.user, password or ""],
            ]
            self._con = jaydebeapi.connect(*connect_args)
        return self._con

    @property
    def owners(self) -> list[str]:
        """Return all owner names.

        Returns:
            list[str]: The list of owners.
        """
        owners_tuples = self.run(
            "SELECT DISTINCT owner FROM systables WHERE owner IS NOT NULL"
        )
        return [owner_tuple[0] for owner_tuple in owners_tuples]

    @property
    def tables(self) -> list[str]:
        """Return all user tables.

        Returns:
            list[str]: The list of tables.
        """
        tables_tuples = self.run(
            "SELECT owner, tabname FROM systables WHERE tabtype = 'T' AND owner IS NOT NULL"
        )
        return [".".join(row) for row in tables_tuples]

    def exists(self, table: str, owner: str | None = None) -> bool:
        """Check whether a table exists.

        Args:
            table (str): The table to check.
            owner (str): The owner to check.

        Returns:
            bool: True if the table exists, False otherwise.
        """
        owner = owner or self.DEFAULT_OWNER
        list_table_info_query = f"""
            SELECT 1
            FROM systables
            WHERE owner = '{owner}' AND tabname = '{table}'
        """  # noqa: S608
        return bool(self.run(list_table_info_query))

    def _check_if_table_exists(self, table: str, owner: str | None = None) -> bool:
        """Check if table exists in a specified owner.

        Args:
            table (str): The table to check.
            owner (str): The owner to check.

        Returns:
            bool: True if the table exists, False otherwise.
        """
        return self.exists(table=table, owner=owner)

    def run(self, query: str) -> list[Record] | bool:
        """Execute a query and return query results.

        Args:
            query (str): The query to execute.

        Returns:
            list[Record] | bool: The result of the query.
        """
        cursor = self.con.cursor()
        cursor.execute(query)

        query_sanitized = query.strip().upper()
        if query_sanitized.startswith("SELECT") or query_sanitized.startswith("WITH"):
            result = cursor.fetchall()
        else:
            result = True

        self.con.commit()
        cursor.close()
        return result

    @add_viadot_metadata_columns
    def to_df(
        self,
        query: str,
        con: Any | None = None,  # noqa: ANN401
        if_empty: Literal["warn", "skip", "fail"] = "warn",
        tests: dict[str, Any] | None = None,
    ) -> pd.DataFrame:
        """Execute a query and return the result as a pandas DataFrame.

        Args:
            query (str): The query to execute.
            con (Any): The connection to use to pull the data.
            if_empty (Literal["warn", "skip", "fail"]):
                What to do if the query returns no data.
            tests (dict[str, Any]): The tests to run on the data.

        Returns:
            pd.DataFrame: The result of the query.
        """
        conn = con or self.con
        query_sanitized = query.strip().upper()

        if query_sanitized.startswith("SELECT") or query_sanitized.startswith("WITH"):
            with conn.cursor() as cursor:
                cursor.execute(query)
                rows = cursor.fetchall()
                cols = (
                    [description[0] for description in cursor.description]
                    if cursor.description
                    else []
                )
            df = pd.DataFrame(rows, columns=cols)
            if df.empty:
                self._handle_if_empty(if_empty=if_empty)
        else:
            self.run(query)
            df = pd.DataFrame()

        if tests:
            validate(df=df, tests=tests)

        return df



"""Eurostat API connector."""

import pandas as pd

from viadot.utils import (
    APIError,
    add_viadot_metadata_columns,
    filter_df_columns,
    handle_api_response,
    validate,
)

from .base import Source


class Eurostat(Source):
    """Eurostat REST API v1 connector.

    This module provides functionalities for connecting to Eurostat  API and download
    the datasets. It includes the following features:
    - Pulling json file with all data from specific dataset.
    - Creating pandas Data Frame from pulled json file.
    - Creating dataset parameters validation if specified.

    Typical usage example:

        eurostat = Eurostat()

        eurostat.to_df(
            dataset_code: str,
            params: dict = None,
            columns: list = None,
            tests: dict = None,
        )
    """

    base_url = "https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data/"

    def __init__(
        self,
        *args,
        dataset_code: str,
        params: dict[str, str] | None = None,
        columns: list[str] | None = None,
        tests: dict | None = None,
        **kwargs,
    ):
        """Initialize the class with Eurostat API data fetching setup.

        This method uses an HTTPS request to pull data from the Eurostat API.
        No API registration or API key is required. Data is fetched based on the
        parameters provided in the dynamic part of the URL.

        Example URL:
        Static part: https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data
        Dynamic part: /TEIBS020/?format=JSON&lang=EN&indic=BS-CSMCI-BAL

        To retrieve specific data, parameters with codes must be provided in `params`.

        Args:
            dataset_code (str):
                The code of the Eurostat dataset to be downloaded.
            params (dict[str, str] | None, optional):
                A dictionary with optional URL parameters. Each key is a parameter ID,
                and the value is a specific parameter code, e.g.,
                `params = {'unit': 'EUR'}` where "unit" is the parameter, and "EUR" is
                the code. Only one code per parameter is allowed. Defaults to None.
            columns (list[str] | None, optional):
                A list of column names (as strings) that are required from the dataset.
                Filters the data to only include the specified columns.
                Defaults to None.
            tests (dict | None, optional):
                A dictionary containing test cases for the data, including:
                - `column_size`: dict{column: size}
                - `column_unique_values`: list[columns]
                - `column_list_to_match`: list[columns]
                - `dataset_row_count`: dict{'min': number, 'max': number}
                - `column_match_regex`: dict{column: 'regex'}
                - `column_sum`: dict{column: {'min': number, 'max': number}}.
                Defaults to None.
            **kwargs:
                Additional arguments passed to the class initializer.
        """
        self.dataset_code = dataset_code
        self.params = params
        self.columns = columns
        self.tests = tests

        super().__init__(*args, **kwargs)

    def get_parameters_codes(self, url: str) -> dict[str, list[str]]:
        """Retrieve available API request parameters and their codes.

        Args:
            url (str): URL built for API call.

        Raises:
            ValueError: If the response from the API is empty or invalid.

        Returns:
            dict[str, list[str]]: Key is a parameter, and value is a list of
                available codes for the specified parameter.
        """
        response = handle_api_response(url)
        data = response.json()

        available_params = data["id"]

        # Dictionary from JSON with keys and related codes values
        dimension = data["dimension"]

        # Assigning list of available codes to specific parameters
        params_and_codes = {}
        for key in available_params:
            if key in dimension:
                codes = list(dimension[key]["category"]["index"].keys())
                params_and_codes[key] = codes
        return params_and_codes

    def validate_params(
        self, dataset_code: str, url: str, params: dict[str, str]
    ) -> None:
        """Validates given parameters against the available parameters in the dataset.

        Important:
            Each dataset in Eurostat has specific parameters that could be used for
            filtering the data. For example dataset ILC_DI04 -Mean and median income by
            household type - EU-SILC and ECHP surveys has parameter such as:
            hhhtyp (Type of household), which can be filtered by specific available
            code of this parameter, such as: 'total', 'a1' (single person),
            'a1_dhc' (single person with dependent children). Please note that each
            dataset has different parameters and different codes

        Raises:
            ValueError: If any of the `params` keys or values is not a string or
            any of them is not available for specific dataset.
        """
        # In order to make params validation, first we need to get params_and_codes.
        key_codes = self.get_parameters_codes(url=url)

        if key_codes is not None:
            # Conversion keys and values on lower cases by using casefold
            key_codes_after_conversion = {
                k.casefold(): [v_elem.casefold() for v_elem in v]
                for k, v in key_codes.items()
            }
            params_after_conversion = {
                k.casefold(): v.casefold() for k, v in params.items()
            }

            # Comparing keys and values
            non_available_keys = [
                key
                for key in params_after_conversion
                if key not in key_codes_after_conversion
            ]
            non_available_codes = [
                value
                for key, value in params_after_conversion.items()
                if key in key_codes_after_conversion
                and value not in key_codes_after_conversion[key]
            ]

            # Error loggers

            if non_available_keys:
                self.logger.error(
                    f"Parameters: '{' | '.join(non_available_keys)}' are not in "
                    "dataset. Please check your spelling!"
                )

                self.logger.error(
                    f"Possible parameters: {' | '.join(key_codes.keys())}"
                )
            if non_available_codes:
                self.logger.error(
                    f"Parameters codes: '{' | '.join(non_available_codes)}' are not "
                    "available. Please check your spelling!\n"
                )
                self.logger.error(
                    "You can find everything via link: https://ec.europa.eu/"
                    f"eurostat/databrowser/view/{dataset_code}/default/table?lang=en"
                )

            if non_available_keys or non_available_codes:
                msg = "Wrong parameters or codes were provided!"
                raise ValueError(msg)

    def eurostat_dictionary_to_df(self, *signals: list[str]) -> pd.DataFrame:
        """Function for creating DataFrame from JSON pulled from Eurostat.

        Returns:
            pd.DataFrame: Pandas DataFrame With 4 columns: index, geo, time, indicator
        """

        class TSignal:
            """Class representing a signal with keys, indexes, labels, and name.

            Attributes:
            signal_keys_list : list[str]
                A list of keys representing unique identifiers for the signal.
            signal_index_list : list[str]
                A list of index values corresponding to the keys of the signal.
            signal_label_list : list[str]
                A list of labels providing human-readable names for the
                signal's keys.
            signal_name : str
                The name of the signal.
            """

            signal_keys_list: list[str]
            signal_index_list: list[str]
            signal_label_list: list[str]
            signal_name: str

        # Dataframe creation
        columns0 = signals[0].copy()
        columns0.append("indicator")
        df = pd.DataFrame(columns=columns0)
        indicator_list = []
        index_list = []
        signal_lists = []

        # Get the dictionary from the inputs
        eurostat_dictionary = signals[-1]

        for signal in signals[0]:
            signal_struct = TSignal()
            signal_struct.signal_name = signal
            signal_struct.signal_keys_list = list(
                eurostat_dictionary["dimension"][signal]["category"]["index"].keys()
            )
            signal_struct.signal_index_list = list(
                eurostat_dictionary["dimension"][signal]["category"]["index"].values()
            )
            signal_label_dict = eurostat_dictionary["dimension"][signal]["category"][
                "label"
            ]
            signal_struct.signal_label_list = [
                signal_label_dict[i] for i in signal_struct.signal_keys_list
            ]
            signal_lists.append(signal_struct)

        col_signal_temp = []
        row_signal_temp = []
        for row_index, row_label in zip(
            signal_lists[0].signal_index_list,
            signal_lists[0].signal_label_list,
            strict=False,
        ):  # rows
            for col_index, col_label in zip(
                signal_lists[1].signal_index_list,
                signal_lists[1].signal_label_list,
                strict=False,
            ):  # cols
                index = str(
                    col_index + row_index * len(signal_lists[1].signal_label_list)
                )
                if index in eurostat_dictionary["value"]:
                    index_list.append(index)
                    col_signal_temp.append(col_label)
                    row_signal_temp.append(row_label)

        indicator_list = [eurostat_dictionary["value"][i] for i in index_list]
        df.indicator = indicator_list
        df[signal_lists[1].signal_name] = col_signal_temp
        df[signal_lists[0].signal_name] = row_signal_temp

        return df

    @add_viadot_metadata_columns
    def to_df(self) -> pd.DataFrame:
        """Function responsible for getting response and creating DataFrame.

        It is using method 'eurostat_dictionary_to_df' with validation
        of provided parameters and their codes if needed.

        Raises:
            TypeError: If self.params is different type than a dictionary.
            TypeError: If self.columns is different type than a list containing strings.
            APIError: If there is an error with the API request.

        Returns:
            pd.DataFrame: Pandas DataFrame.
        """
        # Checking if params and columns have correct type
        if not isinstance(self.params, dict) and self.params is not None:
            msg = "Params should be a dictionary."
            raise TypeError(msg)

        if not isinstance(self.columns, list) and self.columns is not None:
            msg = "Requested columns should be provided as list of strings."
            raise TypeError(msg)

        # Creating url for connection with API
        url = f"{self.base_url}{self.dataset_code}?format=JSON&lang=EN"

        # Making parameters validation
        if self.params is not None:
            self.validate_params(
                dataset_code=self.dataset_code, url=url, params=self.params
            )

        # Getting response from API
        try:
            response = handle_api_response(url, params=self.params)
        except APIError:
            self.validate_params(
                dataset_code=self.dataset_code, url=url, params=self.params
            )
        data = response.json()
        data_frame = self.eurostat_dictionary_to_df(["geo", "time"], data)

        # Merge data_frame with label and last updated date
        label_col = pd.Series(str(data["label"]), index=data_frame.index, name="label")
        last_updated__col = pd.Series(
            str(data["updated"]),
            index=data_frame.index,
            name="updated",
        )
        data_frame = pd.concat([data_frame, label_col, last_updated__col], axis=1)

        # Validation and transformation of requested column
        if self.columns is not None:
            filter_df_columns(data_frame=data_frame, columns=self.columns)

        # Additional validation from utils
        validate(df=data_frame, tests=self.tests)

        return data_frame



from datetime import date, timedelta

from viadot.orchestration.prefect.flows import mindful_to_adls


start_date = date.today() - timedelta(days=2)
end_date = start_date + timedelta(days=1)
date_interval = [start_date, end_date]


def test_mindful_to_adls(mindful_config_key, adls_credentials_secret):
    state = mindful_to_adls(
        azure_key_vault_secret=mindful_config_key,
        endpoint="responses",
        date_interval=date_interval,
        adls_path="raw/dyvenia_sandbox/mindful",
        adls_azure_key_vault_secret=adls_credentials_secret,
        adls_path_overwrite=True,
    )
    assert state.is_successful()



"""Task for downloading data from Vid Club Cloud API."""

from typing import Any, Literal

import pandas as pd
from prefect import task

from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import VidClub


@task(retries=3, log_prints=True, retry_delay_seconds=10, timeout_seconds=2 * 60 * 60)
def vid_club_to_df(  # noqa: PLR0913
    *args: list[Any],
    endpoint: Literal["jobs", "product", "company", "survey"] | None = None,
    from_date: str = "2022-03-22",
    to_date: str | None = None,
    items_per_page: int = 100,
    region: Literal["bg", "hu", "hr", "pl", "ro", "si", "all"] | None = None,
    days_interval: int = 30,
    cols_to_drop: list[str] | None = None,
    azure_key_vault_secret: str | None = None,
    adls_config_key: str | None = None,
    validate_df_dict: dict | None = None,
    timeout: int = 3600,
    **kwargs: dict[str, Any],
) -> pd.DataFrame:
    """Task to downloading data from Vid Club APIs to Pandas DataFrame.

    Args:
        endpoint (Literal["jobs", "product", "company", "survey"], optional):
            The endpoint source to be accessed. Defaults to None.
        from_date (str, optional): Start date for the query, by default is the oldest
            date in the data 2022-03-22.
        to_date (str, optional): End date for the query. By default None,
            which will be executed as datetime.today().strftime("%Y-%m-%d") in code.
        items_per_page (int, optional): Number of entries per page. Defaults to 100.
        region (Literal["bg", "hu", "hr", "pl", "ro", "si", "all"], optional): Region
            filter for the query. Defaults to None (parameter is not used in url).
            [December 2023 status: value 'all' does not work for company and jobs]
        days_interval (int, optional): Days specified in date range per API call
            (test showed that 30-40 is optimal for performance). Defaults to 30.
        cols_to_drop (List[str], optional): List of columns to drop. Defaults to None.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        azure_key_vault_secret (Optional[str], optional): The name of the Azure Key
            Vault secret where credentials are stored. Defaults to None.
        validate_df_dict (dict, optional): A dictionary with optional list of tests
            to verify the output
            dataframe. If defined, triggers the `validate_df` task from task_utils.
            Defaults to None.
        timeout (int, optional): The time (in seconds) to wait while running this task
            before a timeout occurs. Defaults to 3600.

    Returns: Pandas DataFrame
    """
    if not (azure_key_vault_secret or adls_config_key):
        raise MissingSourceCredentialsError

    if not adls_config_key:
        credentials = get_credentials(azure_key_vault_secret)

    vc_obj = VidClub(
        args=args,
        endpoint=endpoint,
        from_date=from_date,
        to_date=to_date,
        items_per_page=items_per_page,
        region=region,
        days_interval=days_interval,
        cols_to_drop=cols_to_drop,
        vid_club_credentials=credentials,
        validate_df_dict=validate_df_dict,
        timeout=timeout,
        kwargs=kwargs,
    )

    return vc_obj.to_df()



"""Tasks for OneStream API."""

from itertools import product
from typing import Any, Literal

import pandas as pd
from prefect import task
import requests

from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources.onestream import OneStream


@task(
    retries=1,
    log_prints=True,
    retry_delay_seconds=5,
    cache_result_in_memory=False,
)
def create_batch_list_of_custom_subst_vars(
    custom_subst_vars: dict[str, list[Any]],
) -> list[dict[str, list[Any]]]:
    """Generate all combinations of custom substitution variables as batch dictionaries.

    Each combination is used as a separate batch when batch_by_subst_vars is True,
    allowing for individual processing and storage of parquet files in S3.

    Args:
        custom_subst_vars (dict[str, list[Any]]): Dictionary where each key maps to a
        list of possible values for that substitution variable. The cartesian product
            of these lists will be computed.

    Returns:
        list[dict[str, list[Any]]]: List of dictionaries for substitution
            variables, each representing a unique combination of custom variable values.
            Each dictionary has the same keys as the input but with single values
                selected from the corresponding lists.

    Raises:
        ValueError: If custom_subst_vars is empty or contains empty lists.
    """
    if not custom_subst_vars:
        msg = "Custom_subst_vars cannot be empty."
        raise ValueError(msg)

    if any(not values for values in custom_subst_vars.values()):
        msg = "All substitution variable lists must contain at least one value."
        raise ValueError(msg)

    return [
        dict(
            zip(
                custom_subst_vars.keys(),
                [[value] for value in combination],
                strict=False,
            )
        )
        for combination in product(*custom_subst_vars.values())
    ]


@task(
    retries=3,
    log_prints=True,
    retry_delay_seconds=10,
    timeout_seconds=60 * 60,
    cache_result_in_memory=False,
)
def onestream_to_df(
    base_url: str,
    application: str,
    api: Literal["data_adapter", "sql_query"],
    credentials_secret: str | None = None,
    config_key: str | None = None,
    params: dict[str, str] | None = None,
    if_empty: Literal["warn", "skip", "fail"] = "fail",
    **kwargs,
) -> pd.DataFrame:
    """Extract data from a OneStream API endpoint into a Pandas DataFrame.

    This is a generic task that works with multiple OneStream endpoint types
    (data_adapter, sql_query) by leveraging the internal dispatcher pattern which
    executes the API endpoint-specific function to get the data.
    All endpoint-specific parameters should be passed as keyword arguments.

    Args:
        base_url (str): The base URL of the OneStream API.
        application (str): The name of the OneStream application.
        api (Literal["data_adapter", "sql_query"]): The API endpoint type.
        credentials_secret (str, optional): Azure Key Vault secret containing
            OneStream credentials. If neither this nor config_key is provided,
            an error is raised. Defaults to None.
        config_key (str, optional): Alternate config key from the local
            environment file. Defaults to None.
        params (dict[str, str], optional): Additional parameters.
            Defaults to None.
        if_empty (Literal["warn", "skip", "fail"], optional): What to do if
            the API returns no data. Defaults to "fail"
        **kwargs: Endpoint-specific parameters.
            For "data_adapter":
                - adapter_name (str): Name of the OneStream adapter to query.
                - workspace_name (str, optional): Name of the workspace.
                  Defaults to "MainWorkspace".
                - adapter_response_key (str, optional): Key in the JSON
                  response. Defaults to "Results".
                - custom_subst_vars (dict[str, list[Any]], optional): Custom
                  substitution variables.
            For "sql_query":
                - sql_query (str): The SQL query to execute.
                - custom_subst_vars (dict, optional): Custom substitution
                  variables.
                - db_location (str, optional): Database location.
                  Defaults to "Application".
                - results_table_name (str, optional): Results table name.
                  Defaults to "Results".
                - external_db (str, optional): External database name.
                  Defaults to "".

    Raises:
        MissingSourceCredentialsError: If neither credentials_secret nor
            config_key is provided.

    Returns:
        pd.DataFrame: The extracted data as a Pandas DataFrame.
    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    credentials = get_credentials(credentials_secret)  # type: ignore
    onestream = OneStream(
        base_url=base_url,
        application=application,
        credentials=credentials,
        config_key=config_key,
        params=params,
    )

    return onestream.to_df(api=api, if_empty=if_empty, **kwargs)


@task(retries=3, log_prints=True, retry_delay_seconds=10, timeout_seconds=60 * 60)
def onestream_run_data_management_seq(
    base_url: str,
    application: str,
    dm_seq_name: str,
    credentials_secret: str | None = None,
    config_key: str | None = None,
    custom_subst_vars: dict[str, list[Any]] | None = None,
    params: dict[str, str] | None = None,
) -> requests.Response:
    """Run a OneStream Data Management Sequence.

    Args:
        base_url (str): OneStream base server URL.
        application (str): OneStream application name.
        dm_seq_name (str): Data Management Sequence name.
        credentials_secret (str, optional): Key Vault secret name. Defaults to None.
        config_key (str,optional): Viadot config key. Defaults to None.
        custom_subst_vars (dict[str, list[Any]], optional): A dictionary mapping
            substitution variable names to lists of possible values.Substition variables
            used as substitution variables that are mapped to the column names in the
            Data Adapter configuration.They provides a kind of SQL WHERE clause for rows
            filtering where variable refers to a name of a set containings a list
            of possible values.
            For example: "prm_activity_region":["MK", "LG"]
                        refers to column UD3 that has custom variable mapping of
                        prm_activity_region.Data will be extracted only for rows where
                        the values (MK,LG) are present.
        params (dict[str, str], optional): API parameters. Defaults to None.

    Returns:
        requests.Response: Sequence execution response.
    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    credentials = get_credentials(credentials_secret)
    onestream = OneStream(
        base_url=base_url,
        application=application,
        credentials=credentials,
        config_key=config_key,
        params=params,
    )

    return onestream.execute(
        api="run_data_management_seq",
        dm_seq_name=dm_seq_name,
        custom_subst_vars=custom_subst_vars,
    )



"""Task for downloading data from ECB exchange rates API to a pandas DataFrame."""

from typing import Any, Literal

import pandas as pd
from prefect import task

from viadot.sources.ecb import ECBExchangeRates


@task(retries=3, log_prints=True, retry_delay_seconds=10, timeout_seconds=60 * 60)
def ecb_to_df(
    if_empty: Literal["warn", "skip", "fail"] = "warn",
    tests: dict[str, Any] | None = None,
    date_filter: list[str] | None = None,
) -> pd.DataFrame:
    """Task to download exchange rates data from ECB API to a pandas DataFrame.

    Args:
        if_empty (Literal["warn", "skip", "fail"], optional): What to do if the
            query returns no data. Defaults to "warn".
        tests (dict[str, Any], optional): A dictionary with optional list of tests
            to verify the output dataframe. If defined, triggers the `validate`
            function from viadot.utils. Defaults to None.

    Examples:
        data_frame = ecb_to_df(if_empty="warn")

    Returns:
        pd.DataFrame: The ECB exchange rates data as a pandas DataFrame with columns:
            - time: The date of the exchange rates
            - currency: The currency code (e.g., USD, GBP)
            - rate: The exchange rate against EUR
    """
    ecb = ECBExchangeRates()

    # Fetch and convert to DataFrame
    return ecb.to_df(
        if_empty=if_empty,
        tests=tests,
        date_filter=date_filter,
    )



from pathlib import Path
import shutil

from loguru import logger

from viadot.orchestration.prefect.tasks import clone_repo


def test_clone_repo_public():
    test_repo_dir = "test_repo_dir"

    assert not Path(test_repo_dir).exists()

    clone_repo.fn(
        url="https://github.com/dbt-labs/dbt-starter-project.git",
        path=test_repo_dir,
        logger=logger,
    )

    assert Path(test_repo_dir).exists()

    shutil.rmtree(test_repo_dir)



"""Download data from a SFTP server to Azure Data Lake Storage."""

from prefect import flow

from viadot.orchestration.prefect.tasks import df_to_adls, sftp_to_df
from viadot.orchestration.prefect.utils import with_flow_timeout_param


@flow(
    name="SFTP extraction to ADLS",
    description="Extract data from a SFTP server and "
    + "load it into Azure Data Lake Storage.",
    retries=1,
    retry_delay_seconds=60,
)
@with_flow_timeout_param()
def sftp_to_adls(
    config_key: str | None = None,
    azure_key_vault_secret: str | None = None,
    file_name: str | None = None,
    sep: str = "\t",
    columns: list[str] | None = None,
    adls_config_key: str | None = None,
    adls_azure_key_vault_secret: str | None = None,
    adls_path: str | None = None,
    adls_path_overwrite: bool = False,
) -> None:
    r"""Flow to download data from a SFTP server to Azure Data Lake.

    Args:
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        azure_key_vault_secret (str, optional): The name of the Azure Key Vault secret
            where credentials are stored. Defaults to None.
        file_name (str, optional): Path to the file in SFTP server. Defaults to None.
        sep (str, optional): The separator to use to read the CSV file.
            Defaults to "\t".
        columns (List[str], optional): Columns to read from the file. Defaults to None.
        adls_config_key (str, optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
        adls_azure_key_vault_secret (str, optional): The name of the Azure Key
            Vault secret containing a dictionary with ACCOUNT_NAME and Service Principal
            credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
            Defaults to None.
        adls_path (str, optional): Azure Data Lake destination file path
            (with file name). Defaults to None.
        adls_path_overwrite (bool, optional): Whether to overwrite the file in ADLS.
            Defaults to True.
    """
    data_frame = sftp_to_df(
        config_key=config_key,
        azure_key_vault_secret=azure_key_vault_secret,
        file_name=file_name,
        sep=sep,
        columns=columns,
    )

    df_to_adls(
        df=data_frame,
        path=adls_path,
        credentials_secret=adls_azure_key_vault_secret,
        config_key=adls_config_key,
        overwrite=adls_path_overwrite,
    )



"""'test_bigquery.py'."""

from viadot.orchestration.prefect.flows import bigquery_to_adls


def test_salesforce_to_adls(azure_key_vault_secret, adls_azure_key_vault_secret):
    """Test SalesForce prefect flow."""
    state = bigquery_to_adls(
        azure_key_vault_secret=azure_key_vault_secret,
        dataset_name="VX_CXU_Data_Feeds",
        table_name="VX_CXU_Operational_CRM_daily_EXPORT",
        adls_path="raw/dyvenia_sandbox/bigquery/bigquery.csv",
        adls_azure_key_vault_secret=adls_azure_key_vault_secret,
        adls_path_overwrite=True,
    )
    all_successful = all(s.type == "COMPLETED" for s in state)
    assert all_successful



"""'test_salesforce.py'."""

from viadot.orchestration.prefect.flows import salesforce_to_adls


def test_salesforce_to_adls(
    azure_key_vault_secret, adls_path, adls_azure_key_vault_secret
):
    """Test Salesforce prefect flow."""
    state = salesforce_to_adls(
        azure_key_vault_secret=azure_key_vault_secret,
        env="dev",
        table="Contact",
        adls_path=adls_path,
        adls_azure_key_vault_secret=adls_azure_key_vault_secret,
        adls_path_overwrite=True,
    )
    all_successful = all(s.type == "COMPLETED" for s in state)
    assert all_successful, "Not all tasks in the flow completed successfully."



"""Source for ingesting data from Matomo API."""

from typing import Any, Literal

import pandas as pd
from pydantic import BaseModel

from viadot.config import get_source_credentials
from viadot.exceptions import APIError
from viadot.sources.base import Source
from viadot.utils import (
    add_viadot_metadata_columns,
    handle_api_response,
    parse_dates,
    validate,
)


HTTP_STATUS_OK = 200


class MatomoCredentials(BaseModel):
    """Check Matomo credentials dict.

    One key is required: api_token_auth.
        - api_token: The unique authentication api_token for the Matomo API.

    Args:
         BaseModel (pydantic.BaseModel): Base class for data validation.
    """

    api_token: str


class Matomo(Source):
    """Matomo source class for fetching data from Matomo API."""

    def __init__(
        self,
        credentials: MatomoCredentials | None = None,
        config_key: str = "matomo",
        *args,
        **kwargs,
    ):
        """Initialize Matomo source with credentials or config key.

        Connector allows to pull data from Matomo API and convert it to a pandas
            DataFrame.

        Args:
            - credentials: MatomoCredentials object or a dict with api_token key.
            - config_key: The key in the viadot config holding relevant credentials.
        Usage:
            matomo = Matomo(credentials=matomo_credentials)

        Example:
            matomo = Matomo(api_token="YOUR_api_token")
            df = matomo.to_df()
        """
        raw_creds = credentials or get_source_credentials(config_key)
        validated_creds = dict(MatomoCredentials(**raw_creds))
        super().__init__(*args, credentials=validated_creds, **kwargs)

    def _validate_parameters(
        self, api_token: str, url: str, params: dict[str, str]
    ) -> dict[str, str]:
        """Validate input parameters and return processed params.

        Args:
            api_token (str): The authentication token for the Matomo API.
            url (str): The base URL of the Matomo instance.
            params (dict[str, str]): Parameters for the API request.

        Returns:
            dict[str, str]: The processed parameters with the api_token added.

        Raises:
            ValueError: If api_token or url are empty, or if required parameters
                are missing.
        """
        if not api_token:
            msg = "api_token is required and cannot be empty"
            self.logger.error(msg)
            raise ValueError(msg)

        if not url:
            msg = "url is required and cannot be empty"
            self.logger.error(msg)
            raise ValueError(msg)

        if not params:
            params = {}

        # Validate essential API parameters
        required_params = ["module", "method", "idSite", "period", "date", "format"]
        missing_params = [
            param
            for param in required_params
            if param not in params or not params[param]
        ]
        if missing_params:
            msg = f"Missing required API parameters: {missing_params}"
            self.logger.error(msg)
            raise ValueError(msg)

        params["token_auth"] = api_token
        return params

    def _validate_response(self, data: dict[str, Any]):
        """Process and validate API response data.

        Args:
            data (dict[str, Any]): The API response data.

        Raises:
            TypeError: If the data is not in expected format (list or dict).
            ValueError: If the data is None.
        """
        if data is None:
            msg = "No data available. Call fetch_data() first."
            self.logger.error(msg)
            raise ValueError(msg)

        # Validate that we have a proper data structure
        if not isinstance(data, list | dict):
            msg = "Data is not in expected format (list or dict)"
            self.logger.error(msg)
            raise TypeError(msg)

    def format_date_range(self, date_range: tuple[str, str]) -> str:
        """Format date range to YYYY-MM-DD,YYYY-MM-DD.

        Args:
            date_range (tuple[str, str]): The date range to format.

        Returns:
            str: The formatted date range.
        """
        date_range_parsed = parse_dates(date_range)
        return f"{date_range_parsed[0]:%Y-%m-%d},{date_range_parsed[1]:%Y-%m-%d}"

    def fetch_data(
        self,
        api_token: str,
        url: str,
        params: dict[str, str],
    ) -> dict[str, Any]:
        """Connect to Matomo API and fetch api response data as JSON.

        The function sends a GET request to the Matomo API endpoint with the specified
        parameters, including the authentication api_token.
        If the request is successful, the response data is stored in the `data`
        attribute of the Matomo instance. If the request fails, an APIError is
        raised with the error message.

        Args:
            api_token (str): The authentication api_token for the Matomo API.
            url (str): The base URL of the Matomo instance.
            params (dict[str, str]): Parameters for the API request.
                Required params: "module","method","idSite","period","date","format".

        Returns:
            dict[str, Any]: The API response data.

        Raises:
            ValueError: If required parameters are missing or invalid.
        """
        params = self._validate_parameters(api_token, url, params)

        try:
            response = handle_api_response(
                url=f"{url}/index.php",
                params=params,
                method="GET",
            )
        except Exception as e:
            msg = "Failed to fetch data from Matomo API"
            self.logger.exception(msg)
            raise APIError(msg) from e

        self.logger.info("Successfully fetched data from Matomo API.")

        data = response.json()
        self._validate_response(data)

        return data

    @add_viadot_metadata_columns
    def to_df(
        self,
        data: dict[str, Any],
        top_level_fields: list[str],
        record_path: str | list[str],
        record_prefix: str | None = None,
        if_empty: Literal["warn", "skip", "fail"] = "warn",
        tests: dict[str, Any] | None = None,
    ) -> pd.DataFrame:
        """Convert Matomo data to pandas DataFrame.

        Args:
            data (dict[str, Any]): The API response data.
            top_level_fields (list[str]): List of top lvl fields to get from
                the response JSON.
            record_path (str or list[str]): The path field to the records in the
                response JSON
                Could be handled as a list of path + fields to extract:
                        record_path = 'actionDetails'
                        record_path = ['actionDetails', 'eventAction']
            record_prefix = A prefix for the record path fields.For example:"action_"
            if_empty: What to do if no data is available.
                     Defaults to "warn".
            tests (Dict[str], optional): A dictionary with optional list of tests
                to verify the output dataframe. If defined, triggers the `validate`
                function from utils. Defaults to None.

        Returns:
            pd.DataFrame: DataFrame containing Matomo data.

        Raises:
            ValueError: If no data has been fetched yet.
        """
        if data is None:
            msg = "No data available. Call fetch_data() first."
            raise ValueError(msg)

        self.logger.info("Converting Matomo data to pandas DataFrame.")

        # Validate record_path exists in data if it's a dict
        if isinstance(data, dict) and record_path not in data:
            self.logger.warning(
                f"record_path '{record_path}' not found in response data."
            )
            self.logger.warning(f"Available keys in response: {list(data.keys())}")

        df = pd.json_normalize(
            data,
            record_path=record_path,
            meta=top_level_fields,
            sep="_",
            errors="ignore",  # ignore puts nan value when field is missing in a record
            record_prefix=record_prefix,
        )

        if df.empty:
            self.logger.warning("No records found in the specified record_path.")
            self._handle_if_empty(if_empty=if_empty)

        if tests:
            self.logger.info("Running data validation tests.")
            validate(df=df, tests=tests)

        self.logger.info(f"Successfully processed {len(df)} records from Matomo data.")

        return df



"""'mediatool.py'."""

import pandas as pd
from prefect import task

from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import Mediatool


@task(retries=3, log_prints=True, retry_delay_seconds=10, timeout_seconds=60 * 60)
def mediatool_to_df(
    config_key: str | None = None,
    azure_key_vault_secret: str | None = None,
    organization_ids: list[str] | None = None,
    media_entries_columns: list[str] | None = None,
) -> pd.DataFrame:
    """Task to download data from Mediatool API.

    Args:
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        azure_key_vault_secret (str, optional): The name of the Azure Key Vault secret
            where credentials are stored. Defaults to None.
        organization_ids (list[str], optional): List of organization IDs.
            Defaults to None.
        media_entries_columns (list[str], optional): Columns to get from media entries.
            Defaults to None.

    Returns:
        pd.DataFrame: The response data as a Pandas Data Frame.
    """
    if not (azure_key_vault_secret or config_key):
        raise MissingSourceCredentialsError

    if not config_key:
        credentials = get_credentials(azure_key_vault_secret)

    mediatool = Mediatool(
        credentials=credentials,
        config_key=config_key,
    )

    return mediatool.to_df(organization_ids, media_entries_columns)



"""Download data from Hubspot API and load into Azure Data Lake Storage."""

from typing import Any

from prefect import flow

from viadot.orchestration.prefect.tasks import df_to_adls, hubspot_to_df
from viadot.orchestration.prefect.utils import with_flow_timeout_param


@flow(
    name="Hubspot extraction to ADLS",
    description="Extract data from Hubspot API and load into Azure Data Lake Storage.",
    retries=1,
    retry_delay_seconds=60,
)
@with_flow_timeout_param()
def hubspot_to_adls(
    config_key: str | None = None,
    azure_key_vault_secret: str | None = None,
    endpoint: str | None = None,
    filters: list[dict[str, Any]] | None = None,
    properties: list[Any] | None = None,
    nrows: int = 1000,
    adls_config_key: str | None = None,
    adls_azure_key_vault_secret: str | None = None,
    adls_path: str | None = None,
    adls_path_overwrite: bool = False,
) -> None:
    """Flow for downloading data from mindful to Azure Data Lake.

    Args:
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        azure_key_vault_secret (Optional[str], optional): The name of the Azure Key
            Vault secret where credentials are stored. Defaults to None.
        endpoint (Optional[str], optional): API endpoint for an individual request.
            Defaults to None.
        filters (Optional[List[Dict[str, Any]]], optional): Filters defined for the API
            body in specific order. Defaults to None.
        properties (Optional[List[Any]], optional): List of user-defined columns to be
            pulled from the API. Defaults to None.
        nrows (int, optional): Max number of rows to pull during execution.
            Defaults to 1000.
        adls_config_key (Optional[str], optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
        adls_azure_key_vault_secret (Optional[str], optional): The name of the Azure Key
            Vault secret containing a dictionary with ACCOUNT_NAME and Service Principal
            credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
            Defaults to None.
        adls_path (Optional[str], optional): Azure Data Lake destination file path
            (with file name). Defaults to None.
        adls_path_overwrite (bool, optional): Whether to overwrite the file in ADLS.
            Defaults to True.

    Examples:
        hubspot_to_adls(
            config_key=config_key,
            endpoint=endpoint,
            nrows=nrows,
            adls_config_key=adls_config_key,
            adls_path=adls_path,
            adls_path_overwrite=True,
        )
    """
    data_frame = hubspot_to_df(
        config_key=config_key,
        azure_key_vault_secret=azure_key_vault_secret,
        endpoint=endpoint,
        filters=filters,
        properties=properties,
        nrows=nrows,
    )

    df_to_adls(
        df=data_frame,
        path=adls_path,
        credentials_secret=adls_azure_key_vault_secret,
        config_key=adls_config_key,
        overwrite=adls_path_overwrite,
    )



"""Test task for pulling data from CloudForCustomers and loading into ADLS."""

from viadot.orchestration.prefect.flows import cloud_for_customers_to_adls


def test_cloud_for_customers_to_adls():
    state = cloud_for_customers_to_adls(
        report_url="https://my341115.crm.ondemand.com/sap/c4c/odata/ana_businessanalytics_analytics.svc/RPZ36A87743F65355C0B904A5QueryResults?$select=TDOC_PRIORITY",
        filter_params={"CBTD_REF_TYPE_CODE": "(%20eq%20%27118%27)"},
        adls_path="raw/c4c/ticket/leads_link/c4c_tickets_leads_link.parquet",
        overwrite=True,
        cloud_for_customers_credentials_secret="aia-c4c-prod",  # noqa: S106
        adls_credentials_secret="app-azure-cr-datalakegen2",  # noqa: S106
    )
    assert state.is_successful()



"""Flows for downloading data from Sharepoint and uploading it to Amazon S3."""

from prefect import flow

from viadot.orchestration.prefect.tasks import s3_upload_file, sharepoint_download_file
from viadot.orchestration.prefect.utils import with_flow_timeout_param


@flow(
    name="extract--sharepoint--s3",
    description="Flows for downloading data from Sharepoint and uploading it to Amazon S3.",
    retries=1,
    retry_delay_seconds=60,
)
@with_flow_timeout_param()
def sharepoint_to_s3(
    url: str,
    local_path: str,
    to_path: str,
    sharepoint_credentials_secret: str | None = None,
    sharepoint_config_key: str | None = None,
    aws_config_key: str | None = None,
) -> None:
    """Download a file from Sharepoint and upload it to S3.

    Args:
        url (str): The URL of the file to be downloaded.
        local_path (str): Local file directory. Defaults to None.
        to_path (str): Where to download the file.
        sharepoint_credentials_secret (str, optional): The name of the secret that
            stores Sharepoint credentials. Defaults to None. More info on:
            https://docs.prefect.io/concepts/blocks/
        sharepoint_config_key (str, optional): The key in the viadot config holding
            relevant credentials.
        aws_config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
    """
    sharepoint_download_file(
        url=url,
        to_path=local_path,
        credentials_secret=sharepoint_credentials_secret,
        config_key=sharepoint_config_key,
    )

    s3_upload_file(
        from_path=local_path,
        to_path=to_path,
        config_key=aws_config_key,
    )



"""Download data from ECB exchange rates API and load into Redshift Spectrum."""

from typing import Any, Literal

from prefect import flow

from viadot.orchestration.prefect.tasks import df_to_redshift_spectrum, ecb_to_df
from viadot.orchestration.prefect.utils import with_flow_timeout_param


@flow(
    name="ECB exchange rates extraction to Redshift Spectrum",
    description="Extract exchange rates data from ECB API and load into Redshift Spectrum.",
    retries=1,
    retry_delay_seconds=60,
)
@with_flow_timeout_param()
def ecb_to_redshift_spectrum(  # noqa: PLR0913
    to_path: str,
    schema_name: str,
    table: str,
    if_empty: Literal["warn", "skip", "fail"] = "warn",
    tests: dict[str, Any] | None = None,
    extension: str = ".parquet",
    if_exists: Literal["overwrite", "append"] = "overwrite",
    partition_cols: list[str] | None = None,
    compression: Literal["snappy", "gzip", "zstd"] | None = None,
    aws_sep: str = ",",
    credentials_secret: str | None = None,
    aws_config_key: str | None = None,
    date_filter: list[str] | None = None,
) -> None:
    """Flow for downloading exchange rates data from ECB API to Redshift Spectrum.

    Args:
        to_path (str): Path to a S3 folder where the table will be located.
        schema_name (str): AWS Glue catalog database name.
        table (str): AWS Glue catalog table name.
        if_empty (Literal["warn", "skip", "fail"], optional): What to do if the
            query returns no data. Defaults to "warn".
        tests (dict[str, Any], optional): A dictionary with optional list of tests
            to verify the output dataframe. If defined, triggers the `validate`
            function from viadot.utils. Defaults to None.
        extension (str): Required file type. Accepted file formats are 'csv' and
            'parquet'. Defaults to ".parquet".
        if_exists (Literal["overwrite", "append"]): Whether to overwrite or append to
            existing data. Defaults to "overwrite".
        partition_cols (list[str], optional): List of column names that will be used to
            create partitions. Defaults to None.
        compression (Literal["snappy", "gzip", "zstd"], optional): Compression style
            (None, snappy, gzip, zstd). Defaults to None.
        aws_sep (str, optional): Field delimiter for the output file. Defaults to ','.
        credentials_secret (str, optional): The name of a secret block in Prefect
            that stores AWS credentials. Defaults to None.
        aws_config_key (str, optional): The key in the viadot config holding relevant
            AWS credentials. Defaults to None.

    Examples:
        ecb_to_redshift_spectrum(
            to_path="s3://your-bucket/ecb/",
            schema_name="staging",
            table="ecb_exchange_rates",
            tests={
                "not_null": ["time", "currency", "rate"],
            },
            partition_cols=["time"],
            compression="snappy",
            if_exists="overwrite",
        )
    """
    data_frame = ecb_to_df(
        if_empty=if_empty,
        tests=tests,
        date_filter=date_filter,
    )

    df_to_redshift_spectrum(
        df=data_frame,
        to_path=to_path,
        schema_name=schema_name,
        table=table,
        extension=extension,
        if_exists=if_exists,
        partition_cols=partition_cols,
        compression=compression,
        sep=aws_sep,
        config_key=aws_config_key,
        credentials_secret=credentials_secret,
    )



from pandas import DataFrame
from pydantic import ValidationError
import pytest

from viadot.orchestration.prefect.tasks.failed_test_email_notification import (
    SmtpConfig,
    find_column,
    find_model,
    find_schema,
    find_test,
)


@pytest.fixture
def failed_dbt_test_result():
    test_dict = {
        "status": "error",
        "timing": [
            {
                "name": "compile",
                "started_at": "2026-01-01T00:00:00.000000Z",
                "completed_at": "2026-01-01T00:00:00.000000Z",
            },
            {
                "name": "execute",
                "started_at": "2026-01-01T00:00:00.000000Z",
                "completed_at": "2026-01-01T00:00:00.000000Z",
            },
        ],
        "thread_id": "Thread-1 (worker)",
        "execution_time": 1.23,
        "message": 'Database Error in test model_test_name (models/gold/model/model.yml)\n  relation "schema.some_table" does not exist',
        "failures": None,
        "unique_id": "test.project.not_null_orders__customer_id.abc123def456",
        "compiled": True,
        "compiled_code": 'SELECT col1 FROM "devdb"."public"."orders" WHERE customer_id IS NOT NULL',
        "relation_name": '"devdb"."public"."orders"',
        "batch_results": None,
        "adapter_response._message": None,
        "adapter_response.rows_affected": None,
        "metadata.generated_at": "2026-01-01T00:00:00.000000Z",
    }

    return DataFrame(test_dict)


def test_find_schema(failed_dbt_test_result):
    result = failed_dbt_test_result["compiled_code"].apply(find_schema)
    assert result[0] == "public"


def test_find_column(failed_dbt_test_result):
    result = failed_dbt_test_result["unique_id"].apply(find_column)
    assert result[0] == "customer_id"


def test_find_model(failed_dbt_test_result):
    result = failed_dbt_test_result["compiled_code"].apply(find_model)
    assert result[0] == "orders"


def test_find_test(failed_dbt_test_result):
    result = failed_dbt_test_result["unique_id"].apply(
        find_test, test_types=["not_null"]
    )
    assert result[0] == "not_null"


def test_smtp_config_defaults():
    config = SmtpConfig(sender="test@gmail.com", password="secret")  # noqa: S106
    assert config.host == "smtp.gmail.com"
    assert config.port == 587


def test_smtp_config_custom_values():
    config = SmtpConfig(
        host="smtp.custom.com",
        port=465,
        sender="test@custom.com",
        password="secret",  # noqa: S106
    )
    assert config.host == "smtp.custom.com"
    assert config.port == 465


def test_smtp_config_missing_password():
    with pytest.raises(ValidationError):
        SmtpConfig(sender="test@gmail.com")  # type: ignore



"""Test Matomo source."""

import json
from pathlib import Path
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest
import requests

from viadot.sources.matomo import Matomo, MatomoCredentials


@pytest.fixture
def sample_matomo_response():
    """Load sample Matomo API response for testing."""
    response_path = Path(__file__).parent / "test_matomo_response.json"
    with response_path.open(encoding="utf-8") as f:
        return json.load(f)


@pytest.fixture
def matomo_credentials():
    """Sample Matomo credentials for testing."""
    return {"api_token": "test_token_123"}


@pytest.fixture
def matomo_instance(matomo_credentials):
    """Create Matomo instance with test credentials."""
    return Matomo(credentials=matomo_credentials)


def test_valid_credentials():
    """Test creating valid MatomoCredentials."""
    creds = MatomoCredentials(api_token="test_token")  # noqa: S106
    assert creds.api_token == "test_token"  # noqa: S105


def test_missing_api_token():
    """Test MatomoCredentials with missing api_token."""
    with pytest.raises(ValueError, match="api_token"):
        MatomoCredentials()


def test_credentials_from_dict():
    """Test creating credentials from dictionary."""
    creds_dict = {"api_token": "dict_token"}
    creds = MatomoCredentials(**creds_dict)
    assert creds.api_token == "dict_token"  # noqa: S105


@patch("viadot.sources.matomo.get_source_credentials")
def test_init_with_credentials_dict(mock_get_creds, matomo_credentials):
    """Test initialization with credentials dictionary."""
    mock_get_creds.return_value = None
    matomo = Matomo(credentials=matomo_credentials)
    assert matomo.credentials == matomo_credentials


@patch("viadot.sources.matomo.get_source_credentials")
def test_init_with_config_key(mock_get_creds, matomo_credentials):
    """Test initialization with config key."""
    mock_get_creds.return_value = matomo_credentials
    matomo = Matomo(config_key="test_matomo")
    assert matomo.credentials == matomo_credentials


@patch("viadot.sources.matomo.get_source_credentials")
def test_init_without_credentials_raises_error(mock_get_creds):
    """Test initialization without credentials raises error."""
    mock_get_creds.return_value = None
    with pytest.raises(TypeError):
        Matomo()


def test_init_sets_data_to_none(matomo_instance):
    """Test that data attribute is initialized to None."""
    assert matomo_instance.data is None


@patch("viadot.sources.matomo.get_source_credentials")
def test_init_with_matomo_credentials_object(mock_get_creds):
    """Test initialization with MatomoCredentials object."""
    mock_get_creds.return_value = None
    creds_obj = MatomoCredentials(api_token="object_token")  # noqa: S106
    # Convert to dict first as the current implementation expects
    creds_dict = creds_obj.dict()
    matomo = Matomo(credentials=creds_dict)
    assert matomo.credentials["api_token"] == "object_token"  # noqa: S105


@patch("viadot.sources.matomo.handle_api_response")
def test_fetch_data_success(
    mock_handle_api_response, matomo_instance, sample_matomo_response
):
    """Test successful data fetching."""
    # Setup mock response
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.json.return_value = sample_matomo_response
    mock_handle_api_response.return_value = mock_response

    # Test parameters
    api_token = "test_token"  # noqa: S105
    url = "https://example.matomo.com"
    params = {
        "module": "API",
        "method": "Live.getLastVisitsDetails",
        "idSite": "1",
        "period": "day",
        "date": "today",
        "format": "JSON",
    }

    # Call fetch_data
    data = matomo_instance.fetch_data(api_token, url, params)

    # Verify the data was stored
    assert data == sample_matomo_response

    # Verify handle_api_response was called correctly
    mock_handle_api_response.assert_called_once()
    call_args = mock_handle_api_response.call_args
    assert call_args[1]["url"] == f"{url}/index.php"
    assert call_args[1]["params"] == {**params, "token_auth": api_token}
    assert call_args[1]["method"] == "GET"


def test_fetch_data_with_empty_params_raises_validation_error(matomo_instance):
    """Test fetch_data with empty params dictionary raises validation error."""
    with pytest.raises(ValueError, match="Missing required API parameters"):
        matomo_instance.fetch_data("token", "https://example.com", {})


def test_fetch_data_with_none_params_raises_validation_error(matomo_instance):
    """Test fetch_data with None params raises validation error."""
    with pytest.raises(ValueError, match="Missing required API parameters"):
        matomo_instance.fetch_data("token", "https://example.com", None)


@patch("viadot.sources.matomo.handle_api_response")
def test_fetch_data_request_exception(mock_handle_api_response, matomo_instance):
    """Test fetch_data with request exception."""
    from viadot.exceptions import APIError

    mock_handle_api_response.side_effect = requests.exceptions.RequestException(
        "Connection failed"
    )

    with pytest.raises(APIError):
        matomo_instance.fetch_data(
            "token",
            "https://example.com",
            {
                "module": "API",
                "method": "test",
                "idSite": "1",
                "period": "day",
                "date": "today",
                "format": "JSON",
            },
        )


def test_fetch_data_missing_api_token(matomo_instance):
    """Test fetch_data raises error when api_token is empty."""
    with pytest.raises(ValueError, match="api_token is required and cannot be empty"):
        matomo_instance.fetch_data("", "https://example.com", {"module": "API"})


def test_fetch_data_missing_url(matomo_instance):
    """Test fetch_data raises error when url is empty."""
    with pytest.raises(ValueError, match="url is required and cannot be empty"):
        matomo_instance.fetch_data("token", "", {"module": "API"})


def test_fetch_data_missing_required_params(matomo_instance):
    """Test fetch_data raises error when required parameters are missing."""
    # Missing all required params
    with pytest.raises(ValueError, match="Missing required API parameters"):
        matomo_instance.fetch_data("token", "https://example.com", {})

    # Missing some required params
    incomplete_params = {
        "module": "API",
        "method": "test",
    }  # missing idSite, period, date, format
    with pytest.raises(ValueError, match="Missing required API parameters"):
        matomo_instance.fetch_data("token", "https://example.com", incomplete_params)


def test_to_df_without_data_raises_error(matomo_instance):
    """Test to_df raises error when no data has been fetched."""
    with pytest.raises(
        ValueError, match="No data available. Call fetch_data\\(\\) first."
    ):
        matomo_instance.to_df(
            data=None, top_level_fields=["test_field"], record_path="actionDetails"
        )


def test_to_df_basic_conversion(matomo_instance, sample_matomo_response):
    """Test basic DataFrame conversion."""
    data = sample_matomo_response
    df = matomo_instance.to_df(
        data=data, top_level_fields=["idSite", "idVisit"], record_path="actionDetails"
    )

    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    # Check that top-level fields are included
    assert "idSite" in df.columns
    assert "idVisit" in df.columns


def test_to_df_with_record_prefix(matomo_instance, sample_matomo_response):
    """Test to_df with record prefix."""
    data = sample_matomo_response

    df = matomo_instance.to_df(
        data=data,
        top_level_fields=["idSite"],
        record_path="actionDetails",
        record_prefix="action_",
    )

    # Check that actionDetails fields have the prefix
    action_columns = [col for col in df.columns if col.startswith("action_")]
    assert len(action_columns) != 0


def test_to_df_with_list_record_path(matomo_instance, sample_matomo_response):
    """Test to_df with list record_path."""
    data = sample_matomo_response

    df = matomo_instance.to_df(
        data=data, top_level_fields=["idSite"], record_path=["actionDetails"]
    )

    assert isinstance(df, pd.DataFrame)
    assert not df.empty


def test_to_df_empty_result_warn(matomo_instance):
    """Test to_df with empty result and warn if_empty."""
    data = {"actionDetails": []}

    with patch("viadot.sources.base.logger.warning") as mock_warning:
        df = matomo_instance.to_df(
            data=data,
            top_level_fields=["idSite"],
            record_path="actionDetails",
            if_empty="warn",
        )

        assert df.empty
        # Warning is called twice: once in to_df, once in _handle_if_empty
        assert mock_warning.call_count == 2
        mock_warning.assert_any_call("No records found in the specified record_path.")
        mock_warning.assert_any_call("The query produced no data.")


def test_to_df_empty_result_skip(matomo_instance):
    """Test to_df with empty result and skip if_empty."""
    from viadot.signals import SKIP

    data = {"actionDetails": []}

    with pytest.raises(SKIP):
        matomo_instance.to_df(
            data=data,
            top_level_fields=["idSite"],
            record_path="actionDetails",
            if_empty="skip",
        )


def test_to_df_empty_result_fail(matomo_instance):
    """Test to_df with empty result and fail if_empty."""
    data = {"actionDetails": []}

    with pytest.raises(ValueError, match="The query produced no data"):
        matomo_instance.to_df(
            data=data,
            top_level_fields=["idSite"],
            record_path="actionDetails",
            if_empty="fail",
        )


def test_to_df_with_tests_validation(matomo_instance, sample_matomo_response):
    """Test to_df with tests validation."""
    data = sample_matomo_response

    test_config = {"column_tests": {"idSite": {"not_null": True}}}

    with patch("viadot.sources.matomo.validate") as mock_validate:
        df = matomo_instance.to_df(
            data=data,
            top_level_fields=["idSite"],
            record_path="actionDetails",
            tests=test_config,
        )

        mock_validate.assert_called_once()
        args, kwargs = mock_validate.call_args
        assert kwargs["df"] is df
        assert kwargs["tests"] == test_config


@patch("viadot.sources.matomo.handle_api_response")
def test_full_workflow(
    mock_handle_api_response, matomo_instance, sample_matomo_response
):
    """Test complete workflow from fetch to DataFrame."""
    # Setup mock response
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.json.return_value = sample_matomo_response
    mock_handle_api_response.return_value = mock_response

    # Fetch data
    matomo_instance.fetch_data(
        "test_token",
        "https://example.matomo.com",
        {
            "module": "API",
            "method": "Live.getLastVisitsDetails",
            "idSite": "1",
            "period": "day",
            "date": "today",
            "format": "JSON",
        },
    )

    # Convert to DataFrame
    df = matomo_instance.to_df(
        data=sample_matomo_response,
        top_level_fields=["idSite", "idVisit"],
        record_path="actionDetails",
    )

    # Verify results
    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert "idSite" in df.columns
    assert "idVisit" in df.columns


def test_data_persistence(matomo_instance, sample_matomo_response):
    """Test that data persists between method calls."""
    # First call
    df1 = matomo_instance.to_df(
        data=sample_matomo_response,
        top_level_fields=["idSite"],
        record_path="actionDetails",
    )
    # Second call
    df2 = matomo_instance.to_df(
        data=sample_matomo_response,
        top_level_fields=["idVisit"],
        record_path="actionDetails",
    )

    assert not df1.empty
    assert not df2.empty



"""SAP RFC connectors."""

from collections import OrderedDict
from collections import OrderedDict as OrderedDictType
from collections.abc import Iterable, Iterator
import logging
import re
from typing import (
    Any,
    Literal,
)

import numpy as np
from numpy.typing import ArrayLike
import pandas as pd
import pyrfc


try:
    import pyrfc
    from pyrfc._exception import ABAPApplicationError
except ModuleNotFoundError as e:
    msg = "Missing required modules to use SAPRFC source."
    raise ImportError(msg) from e

from sql_metadata import Parser

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError, DataBufferExceededError
from viadot.orchestration.prefect.utils import DynamicDateHandler
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns, validate


logger = logging.getLogger()


def adjust_where_condition_by_adding_missing_spaces(sql: str) -> str:
    """Add white spaces between operators and `WHERE` statement.

    This function is taking raw sql string and sanitizing it at the beginning of the
    'query()' method, so other methods that taking sql as parameter could have sql
    without whitespaces issues.

    Args:
        sql (str): raw sql query passed in flow

    Returns:
        str: sql query after adding white spaces if needed
    """
    # Check if 'WHERE' statement is not attached to 'FROM' or column name as there is
    # a need for space on both sides of 'WHERE'.
    sql = re.sub(rf"{re.escape('WHERE')}(?<!\s)", "WHERE ", sql, flags=re.IGNORECASE)
    sql = re.sub(rf"(?<!\s){re.escape('WHERE')}", " WHERE", sql, flags=re.IGNORECASE)
    sql = re.sub(r"\s+", " ", sql)

    # Check if operators are not attached to column or value as there is need for space
    # on both sides of operator.
    operators = ["<>", "!=", "<=", ">=", "!<", "!>", "=", ">", "<"]
    reverse_check = [
        "< >",
        "! =",
        "< =",
        "> =",
        "! <",
        "! >",
    ]

    for op in operators:
        sql = re.sub(rf"(?<!\s){re.escape(op)}", f" {op}", sql)
        sql = re.sub(rf"{re.escape(op)}(?<!\s)", f"{op} ", sql)
        sql = re.sub(r"\s+", " ", sql)
    for op_2 in reverse_check:
        if op_2 in sql:
            sql = sql.replace(op_2, "".join(op_2.split()))

    return sql


def _remove_whitespaces(text: str) -> str:
    return " ".join(text.split())


def _get_keyword_for_condition(where: str, condition: str) -> str:
    where = where[: where.find(condition)]
    return where.split()[-1]


def _get_where_uppercased(where: str) -> str:
    """Uppercase a WHERE clause's keywords without altering the original string."""
    where_and_uppercased = re.sub("\\sand ", " AND ", where)
    return re.sub("\\sor ", " OR ", where_and_uppercased)


def _remove_last_condition(where: str) -> str:
    """Remove the last condtion from a WHERE clause."""
    where = _get_where_uppercased(where)
    split_by_and = re.split("\\sAND ", where)
    conditions = [re.split("\\sOR ", expr) for expr in split_by_and]
    conditions_flattened = [
        condition for sublist in conditions for condition in sublist
    ]

    condition_to_remove = conditions_flattened[-1]

    where_trimmed = where[: where.find(condition_to_remove)].split()
    where_trimmed_without_last_keyword = " ".join(where_trimmed[:-1])

    return where_trimmed_without_last_keyword, condition_to_remove


def _trim_where(where: str) -> tuple[str, OrderedDictType[str, str] | None]:
    """Trim a WHERE clause to 75 characters or less, as required by SAP RFC.

    The rest of filters will be applied in-memory on client side.
    """
    if len(where) <= SAPRFC.COL_CHARACTER_WIDTH_LIMIT:
        return where, None

    wheres_to_add = OrderedDict()
    keywords_with_conditions = []
    where_trimmed = where
    while len(where_trimmed) > SAPRFC.COL_CHARACTER_WIDTH_LIMIT:
        # trim the where
        where_trimmed, removed_condition = _remove_last_condition(where_trimmed)

        # store the removed conditions so we can readd them later
        keyword = _get_keyword_for_condition(where, removed_condition)
        keywords_with_conditions.append((keyword, removed_condition))

    wheres_to_add_sorted = keywords_with_conditions[::-1]
    wheres_to_add = OrderedDict(wheres_to_add_sorted)

    return where_trimmed, wheres_to_add


def _detect_extra_rows(
    row_index: int,
    data_raw: ArrayLike,
    chunk: int,
    fields: list[str],
) -> int | ArrayLike | bool:
    """Check if, in between calls to the SAP table, the number of rows has increased.

    If so, remove the last rows added, to fit the size of the previous columns.

    Args:
        row_index (int): Number of rows set it down in he first iteration with the SAP
            table.
        data_raw (ArrayLike): Array with the data retrieve from SAP table.
        chunk (int): The part number in which a number of SAP table columns have been
            split.
        fields (List[str]): A list with the names of the columns in a chunk.

    Returns:
        Union[int, ArrayLike, bool]: A tuple with the parameters "row_index",
            "data_raw", a new boolean variable "start" to indicate when the for loop has
            to be restarted, and "chunk" variable.
    """
    start = False
    if row_index == 0:
        row_index = data_raw.shape[0]
        if row_index == 0:
            logger.warning(
                f"Empty output was generated for chunk {chunk} in columns {fields}."
            )
            start = True
    elif data_raw.shape[0] != row_index:
        data_raw = data_raw[:row_index]
        logger.warning(
            "New rows were generated during the execution of the script. The table is truncated to the number of rows for the first chunk."
        )

    return row_index, data_raw, start


def _gen_split(data: Iterable[str], sep: str, record_key: str) -> Iterator[list[str]]:
    """Split each string in the given iterable using the specified separator.

    Helps to reduce memory usage when processing big data sets by yielding the resulting
    list.

    Args:
        data: An iterable collection of strings to be split.
        sep: The string separator used to split each string in `data`.
        record_key: A key for extraction of the data from the nested dict.

    Yields:
        A list of substrings for each string in `data`, split using `sep`.

    Examples:
        >>> list(gen_strip(["a|b|c", "d|e|f|g"], "|", "WA"))
        [['a', 'b', 'c'], ['d', 'e', 'f', 'g']]
    """
    for row in data:
        yield row[record_key].split(sep)


class SAPRFC(Source):
    """A class for querying SAP with SQL using the RFC protocol.

    Note that only a very limited subset of SQL is supported:
    - aliases
    - where clauses combined using the AND operator
    - limit & offset

    Unsupported:
    - aggregations
    - joins
    - subqueries
    - etc.
    """

    COL_CHARACTER_WIDTH_LIMIT = 75

    def __init__(
        self,
        sep: str | None = "♔",
        replacement: str = "-",
        func: str = "RFC_READ_TABLE",
        rfc_total_col_width_character_limit: int = 400,
        rfc_unique_id: list[str] | None = None,
        credentials: dict[str, Any] | None = None,
        config_key: str | None = None,
        *args,
        **kwargs,
    ):
        """Create an instance of the SAPRFC class.

        Args:
            sep (str, optional): The separator to use when reading query results.
                If set to None, multiple options are automatically tried.
                Defaults to ♔.
            replacement (str, optional): In case of separator is on a columns, set up a
                new character to replace inside the string to avoid flow breakdowns.
                Defaults to "-".
            func (str, optional): SAP RFC function to use. Defaults to "RFC_READ_TABLE".
            rfc_total_col_width_character_limit (int, optional): Number of characters by
                which query will be split in chunks in case of too many columns for RFC
                function. According to SAP documentation, the limit is 512 characters.
                However, we observed SAP raising an exception even on a slightly lower
                number of characters, so we add a safety margin. Defaults to 400.
            rfc_unique_id  (List[str], optional): Reference columns to merge chunks
                DataFrames. These columns must to be unique. Defaults to None.
            credentials (Dict[str, Any], optional): 'api_key'. Defaults to None.
            config_key (str, optional): The key in the viadot config holding relevant
                credentials.

        Raises:
            CredentialError: If provided credentials are incorrect.
        """
        self._con = None

        credentials = credentials or get_source_credentials(config_key)
        if credentials is None:
            msg = "Please specify the credentials."
            raise CredentialError(msg)

        super().__init__(*args, credentials=credentials, **kwargs)

        self.sep = sep
        self.replacement = replacement
        self.client_side_filters = None
        self.func = func
        self.rfc_total_col_width_character_limit = rfc_total_col_width_character_limit

        if rfc_unique_id is not None:
            self.rfc_unique_id = list(set(rfc_unique_id))
            self._rfc_unique_id_len = {}
        else:
            self.rfc_unique_id = None

    @property
    def con(self) -> pyrfc.Connection:
        """The pyRFC connection to SAP."""
        if self._con is not None:
            return self._con
        con = pyrfc.Connection(**self.credentials)
        self._con = con
        return con

    def check_connection(self) -> None:
        """Check the connection to SAP."""
        self.logger.info("Checking the connection...")
        self.con.ping()
        self.logger.info("Connection has been validated successfully.")

    def close_connection(self) -> None:
        """Close the SAP RFC connection."""
        self.con.close()
        self.logger.info("Connection has been closed successfully.")

    def get_function_parameters(
        self,
        function_name: str,
        description: None | Literal["short", "long"] = "short",
        *args,
    ) -> list[str] | pd.DataFrame:
        """Get the description for a SAP RFC function.

        Args:
            function_name (str): The name of the function to detail.
            description (Union[None, Literal[, optional): Whether to display
            a short or a long description. Defaults to "short".

        Raises:
            ValueError: If the argument for description is incorrect.

        Returns:
            Union[List[str], pd.DataFrame]: Either a list of the function's
            parameter names (if 'description' is set to None),
            or a short or long description.
        """
        if description not in ["short", "long"]:
            msg = "Incorrect value for 'description'. Correct values: None, 'short', 'long'."
            raise ValueError(msg)

        descr = self.con.get_function_description(function_name, *args)
        param_names = [param["name"] for param in descr.parameters]
        detailed_params = descr.parameters
        filtered_detailed_params = [
            {
                "name": param["name"],
                "parameter_type": param["parameter_type"],
                "default_value": param["default_value"],
                "optional": param["optional"],
                "parameter_text": param["parameter_text"],
            }
            for param in descr.parameters
        ]

        if description is not None:
            if description == "long":
                params = detailed_params
            else:
                params = filtered_detailed_params
            params = pd.DataFrame.from_records(params)
        else:
            params = param_names

        return params

    def _get_where_condition(self, sql: str) -> str:
        """Retrieve the WHERE conditions from a SQL query.

        Args:
            sql (str): The input SQL query.

        Raises:
            ValueError: Raised if the WHERE clause is longer than
            75 characters (SAP's limitation) and the condition for the
            extra clause(s) is OR.

        Returns:
            str: The where clause trimmed to <= 75 characters.
        """
        where_match = re.search("\\sWHERE ", sql.upper())
        if not where_match:
            return None

        limit_match = re.search("\\sLIMIT ", sql.upper())
        limit_pos = limit_match.span()[0] if limit_match else len(sql)

        where = sql[where_match.span()[1] : limit_pos]
        where_sanitized = _remove_whitespaces(where)
        where_trimmed, client_side_filters = _trim_where(where_sanitized)
        if client_side_filters:
            self.logger.warning(
                "A WHERE clause longer than 75 character limit detected."
            )
            if "OR" in [key.upper() for key in client_side_filters]:
                msg = "WHERE conditions after the 75 character limit can only be combined with the AND keyword."
                raise ValueError(msg)
            for val in client_side_filters.values():
                if ")" in val:
                    msg = "Nested conditions eg. AND (col_1 = 'a' AND col_2 = 'b') found between or after 75 characters in WHERE condition!"
                    msg += " Please change nested conditions part of query separated with 'AND' keywords,"
                    msg += " or place nested conditions part at the beginning of the where statement."
                    raise ValueError(msg)
            filters_pretty = list(client_side_filters.items())
            self.logger.warning(
                f"Trimmed conditions ({filters_pretty}) will be applied client-side."
            )
            self.logger.warning("See the documentation for caveats.")

        self.client_side_filters = client_side_filters
        return where_trimmed

    @staticmethod
    def _get_table_name(sql: str) -> str:
        parsed = Parser(sql)
        if len(parsed.tables) > 1:
            msg = "Querying more than one table is not supported."
            raise ValueError(msg)
        return parsed.tables[0]

    def _build_pandas_filter_query(
        self, client_side_filters: OrderedDictType[str, str]
    ) -> str:
        """Build a WHERE clause that will be applied client-side.

        This is required if the WHERE clause passed to query() is
        longer than 75 characters.

        Args:
            client_side_filters (OrderedDictType[str, str]): The
            client-side filters to apply.

        Returns:
            str: the WHERE clause reformatted to fit the format
            required by DataFrame.query().
        """
        for i, f in enumerate(client_side_filters.items()):
            if i == 0:
                # skip the first keyword; we assume it's "AND"
                query = f[1]
            else:
                query += " " + f[0] + " " + f[1]

            filter_column_name = f[1].split()[0]
            resolved_column_name = self._resolve_col_name(filter_column_name)
        return re.sub("\\s?=\\s?", " == ", query).replace(
            filter_column_name, resolved_column_name
        )

    def extract_values(self, sql: str) -> None:
        """TODO: This should cover all values, not just columns."""
        self.where = self._get_where_condition(sql)
        self.select_columns = self._get_columns(sql, aliased=False)
        self.select_columns_aliased = self._get_columns(sql, aliased=True)

    def _resolve_col_name(self, column: str) -> str:
        """Get aliased column name if it exists, otherwise return column name."""
        return self.aliases_keyed_by_columns.get(column, column)

    def _get_columns(self, sql: str, aliased: bool = False) -> list[str]:
        """Retrieve column names from a SQL query.

        Args:
            sql (str): The SQL query to parse.
            aliased (bool, optional): Whether to returned aliased
            names. Defaults to False.

        Returns:
            List[str]: A list of column names.
        """
        parsed = Parser(sql)
        columns = list(parsed.columns_dict["select"])
        if aliased:
            aliases_keyed_by_alias = parsed.columns_aliases
            aliases_keyed_by_columns = OrderedDict(
                {val: key for key, val in aliases_keyed_by_alias.items()}
            )

            self.aliases_keyed_by_columns = aliases_keyed_by_columns

            columns = [aliases_keyed_by_columns.get(col, col) for col in columns]

        if self.client_side_filters:
            # In case the WHERE clause is > 75 characters long, we execute the rest of
            # the filters client-side. To do this, we need to pull all fields in the
            # client-side WHERE conditions. Below code adds these columns to the list of
            # SELECTed fields.
            cols_to_add = [v.split()[0] for v in self.client_side_filters.values()]
            if aliased:
                cols_to_add = [aliases_keyed_by_columns[col] for col in cols_to_add]
            columns.extend(cols_to_add)
            columns = list(dict.fromkeys(columns))  # remove duplicates

        return columns

    @staticmethod
    def _get_limit(sql: str) -> int | None:
        """Get limit from the query."""
        limit_match = re.search("\\sLIMIT ", sql.upper())
        if not limit_match:
            return None

        return int(sql[limit_match.span()[1] :].split()[0])

    @staticmethod
    def _get_offset(sql: str) -> int | None:
        """Get offset from the query."""
        offset_match = re.search("\\sOFFSET ", sql.upper())
        if not offset_match:
            return None

        return int(sql[offset_match.span()[1] :].split()[0])

    def _parse_dates(
        self,
        query: str,
        dynamic_date_symbols: list[str] = ["<<", ">>"],  # noqa: B006
        dynamic_date_format: str = "%Y%m%d",
        dynamic_date_timezone: str = "UTC",
    ) -> str:
        """Process dynamic dates inside the query and validate used patterns type.

        Args:
            query (str): The SQL query to be processed.
            dynamic_date_symbols (list[str], optional): Symbols used for dynamic date
                handling. Defaults to ["<<", ">>"].
            dynamic_date_format (str, optional): Format used for dynamic date parsing.
                Defaults to "%Y%m%d".
            dynamic_date_timezone (str, optional): Timezone used for dynamic date
                processing. Defaults to "UTC".

        Returns:
            str: The processed SQL query with dynamic dates replaced.

        Raises:
            TypeError: If the query contains dynamic date patterns that generate
        a range of dates.
        """
        ddh = DynamicDateHandler(
            dynamic_date_symbols=dynamic_date_symbols,
            dynamic_date_format=dynamic_date_format,
            dynamic_date_timezone=dynamic_date_timezone,
        )
        processed_sql_or_list = ddh.process_dates(query)

        if isinstance(processed_sql_or_list, list):
            msg = (
                f"This query contains {ddh._find_dynamic_date_patterns(query)} dynamic date(s) "
                "that generate a range of dates, which is currently not supported"
                "in query generation. Please use one of the singular pattern dynamic date symbols."
            )
            raise TypeError(msg)

        return processed_sql_or_list

    # Holy crap what a mess. TODO: refactor this so it can be even remotely tested...
    def query(self, sql: str, sep: str | None = None) -> None:  # noqa: C901, PLR0912, PLR0915
        """Parse an SQL query into pyRFC commands and save it into an internal dict.

        Args:
            sql (str): The SQL query to be ran.
            sep (str, optional): The separator to be used
            to split columns in the result blob. Defaults to self.sep.

        Raises:
            ValueError: If the query is not a SELECT query.
        """
        if not sql.strip().upper().startswith("SELECT"):
            msg = "Only SELECT queries are supported."
            raise ValueError(msg)

        sep = sep if sep is not None else self.sep

        sql = adjust_where_condition_by_adding_missing_spaces(sql=sql)
        self.sql = sql

        self.extract_values(sql)

        table_name = self._get_table_name(sql)
        # this has to be called before checking client_side_filters
        where = self.where
        columns = self.select_columns
        lists_of_columns = []
        cols = []
        col_length_total = 0
        if isinstance(self.rfc_unique_id, list):
            character_limit = self.rfc_total_col_width_character_limit
            for rfc_unique_col in self.rfc_unique_id:
                rfc_unique_col_len = int(
                    self.call(
                        "DDIF_FIELDINFO_GET",
                        TABNAME=table_name,
                        FIELDNAME=rfc_unique_col,
                    )["DFIES_TAB"][0]["LENG"]
                )
                if rfc_unique_col_len > int(
                    self.rfc_total_col_width_character_limit / 4
                ):
                    msg = f"{rfc_unique_col} can't be used as unique column, too large."
                    raise ValueError(msg)
                local_limit = (
                    self.rfc_total_col_width_character_limit - rfc_unique_col_len
                )
                character_limit = min(local_limit, character_limit)
                self._rfc_unique_id_len[rfc_unique_col] = rfc_unique_col_len
        else:
            character_limit = self.rfc_total_col_width_character_limit

        for col in columns:
            info = self.call("DDIF_FIELDINFO_GET", TABNAME=table_name, FIELDNAME=col)
            col_length = info["DFIES_TAB"][0]["LENG"]
            col_length_total += int(col_length)
            if col_length_total <= character_limit:
                cols.append(col)
            else:
                if isinstance(self.rfc_unique_id, list):
                    if all(
                        rfc_unique_col not in cols
                        for rfc_unique_col in self.rfc_unique_id
                    ):
                        for rfc_unique_col in self.rfc_unique_id:
                            if rfc_unique_col not in cols:
                                cols.append(rfc_unique_col)
                else:
                    msg = "Columns character limit per row exceeded. Please select fewer columns or add rfc_unique_id."
                    raise DataBufferExceededError(msg)
                lists_of_columns.append(cols)
                cols = [col]
                col_length_total = int(col_length)

        if isinstance(self.rfc_unique_id, list) and all(
            rfc_unique_col not in cols for rfc_col in self.rfc_unique_id
        ):
            for rfc_unique_col in self.rfc_unique_id:
                if rfc_unique_col not in cols:
                    cols.append(rfc_unique_col)
        lists_of_columns.append(cols)

        columns = lists_of_columns
        options = [{"TEXT": where}] if where else None
        limit = self._get_limit(sql)
        offset = self._get_offset(sql)
        query_json = {
            "QUERY_TABLE": table_name,
            "FIELDS": columns,
            "OPTIONS": options,
            "ROWCOUNT": limit,
            "ROWSKIPS": offset,
            "DELIMITER": sep,
        }
        # SAP doesn't understand None, so we filter out non-specified parameters
        query_json_filtered = {
            key: query_json[key] for key in query_json if query_json[key] is not None
        }
        self._query = query_json_filtered

    def call(self, func: str, *args, **kwargs) -> dict[str, Any]:
        """Call a SAP RFC function."""
        return self.con.call(func, *args, **kwargs)

    def _get_alias(self, column: str) -> str:
        return self.aliases_keyed_by_columns.get(column, column)

    def _get_client_side_filter_cols(self):
        return [f[1].split()[0] for f in self.client_side_filters.items()]

    def _adjust_whitespaces(self, df: pd.DataFrame) -> pd.DataFrame:
        """Adjust the number of whitespaces.

        Add whitespace characters in each row of each unique column to achieve
        equal length of values in these columns, ensuring proper merging of subqueries.

        """
        for rfc_unique_col in self.rfc_unique_id:
            # Check in SAP metadata what is the declared
            # dtype characters amount
            rfc_unique_column_len = self._rfc_unique_id_len[rfc_unique_col]
            actual_length_of_field = df[rfc_unique_col].str.len()
            # Check which rows have fewer characters
            # than specified in the column data type.
            rows_missing_whitespaces = actual_length_of_field < rfc_unique_column_len
            if any(rows_missing_whitespaces):
                # Check how many whitespaces are missing in each row.
                logger.info(f"Adding whitespaces for {rfc_unique_col} column")
                n_missing_whitespaces = rfc_unique_column_len - actual_length_of_field
                df.loc[rows_missing_whitespaces, rfc_unique_col] += np.char.multiply(
                    " ", n_missing_whitespaces[rows_missing_whitespaces]
                )
        return df

    # TODO: refactor to remove linter warnings and so this can be tested.
    @add_viadot_metadata_columns
    def to_df(self, tests: dict | None = None) -> pd.DataFrame:  # noqa: C901, PLR0912, PLR0915
        """Load the results of a query into a pandas DataFrame.

        Due to SAP limitations, if the length of the WHERE clause is longer than 75
        characters, we trim whe WHERE clause and perform the rest of the filtering
        on the resulting DataFrame. Eg. if the WHERE clause contains 4 conditions
        and has 80 characters, we only perform 3 filters in the query, and perform
        the last filter on the DataFrame. If characters per row limit will be exceeded,
        data will be downloaded in chunks.

        Source: https://success.jitterbit.com/display/DOC/Guide+to+Using+RFC_READ_TABLE+to+Query+SAP+Tables#GuidetoUsingRFC_READ_TABLEtoQuerySAPTables-create-the-operation
        - WHERE clause: 75 character limit
        - SELECT: 512 character row limit

        Args:
            tests (Dict[str], optional): A dictionary with optional list of tests
                to verify the output dataframe. If defined, triggers the `validate`
                function from utils. Defaults to None.

        Returns:
            pd.DataFrame: A DataFrame representing the result of the query provided in
                `PyRFC.query()`.
        """
        params = self._query
        sep = self._query.get("DELIMITER")
        fields_lists = self._query.get("FIELDS")
        if len(fields_lists) > 1:
            logger.info(f"Data will be downloaded in {len(fields_lists)} chunks.")
        func = self.func
        if sep is None:
            # Automatically find a working separator.
            separators = [
                "|",
                "/t",
                "#",
                ";",
                "@",
                "%",
                "^",
                "`",
                "~",
                "{",
                "}",
                "$",
            ]
        else:
            separators = [sep]

        for sep in separators:
            logger.info(f"Checking if separator '{sep}' works.")
            if isinstance(self.rfc_unique_id, list):
                # Columns only for the first chunk. We add the rest later to avoid name
                # conflicts.
                df = pd.DataFrame(columns=fields_lists[0])
            else:
                df = pd.DataFrame()
            self._query["DELIMITER"] = sep
            chunk = 1
            row_index = 0
            for fields in fields_lists:
                logger.info(f"Downloading {chunk} data chunk...")
                self._query["FIELDS"] = fields
                try:
                    response = self.call(func, **params)
                except ABAPApplicationError as e:
                    if e.key == "DATA_BUFFER_EXCEEDED":
                        msg = "Character limit per row exceeded. Please select fewer columns."
                        raise DataBufferExceededError(msg) from e
                    raise
                # Check and skip if there is no data returned.
                if response["DATA"]:
                    record_key = "WA"
                    data_raw = np.array(response["DATA"])
                    del response
                    # If reference columns are provided, it's not necessary to remove
                    # any extra row.
                    if not isinstance(self.rfc_unique_id, list):
                        row_index, data_raw, start = _detect_extra_rows(
                            row_index, data_raw, chunk, fields
                        )
                    else:
                        start = False
                    records = list(_gen_split(data_raw, sep, record_key))
                    del data_raw
                    if (
                        isinstance(self.rfc_unique_id, list)
                        and list(df.columns) != fields
                    ):
                        df_tmp = pd.DataFrame(columns=fields)
                        df_tmp[fields] = records
                        df_tmp = self._adjust_whitespaces(df_tmp)
                        df = pd.merge(df, df_tmp, on=self.rfc_unique_id, how="outer")
                    elif not start:
                        df[fields] = records
                    else:
                        df[fields] = np.nan
                    chunk += 1
                elif not response["DATA"]:
                    logger.warning("No data returned from SAP.")
        if not df.empty:
            # It is used to filter out columns which are not in select query
            # for example columns passed only as unique column
            df = df.loc[:, self.select_columns]
            df.columns = self.select_columns_aliased

        if self.client_side_filters:
            filter_query = self._build_pandas_filter_query(self.client_side_filters)
            df.query(filter_query, inplace=True)
            client_side_filter_cols_aliased = [
                self._get_alias(col) for col in self._get_client_side_filter_cols()
            ]
            cols_to_drop = [
                col
                for col in client_side_filter_cols_aliased
                if col not in self.select_columns_aliased
            ]
            df.drop(cols_to_drop, axis=1, inplace=True)
        self.close_connection()

        if tests:
            validate(df=df, tests=tests)

        return df



"""Flows for downloading data from SQLServer and uploading it to Redshift Spectrum."""

from typing import Any, Literal

from prefect import flow

from viadot.orchestration.prefect.tasks import df_to_redshift_spectrum, sql_server_to_df
from viadot.orchestration.prefect.utils import (
    DynamicDateHandler,
    with_flow_timeout_param,
)


@flow(
    name="extract--sql_server--redshift_spectrum",
    description="Extract data from SQLServer and load it into AWS Redshift Spectrum.",
    retries=1,
    retry_delay_seconds=60,
)
@with_flow_timeout_param()
def sql_server_to_redshift_spectrum(  # noqa: PLR0913
    query: str,
    to_path: str,
    schema_name: str,
    table: str,
    sep: str = ",",
    extension: str = ".parquet",
    if_exists: Literal["overwrite", "append"] = "overwrite",
    partition_cols: list[str] | None = None,
    index: bool = False,
    compression: str | None = None,
    dynamic_date_symbols: list[str] = ["<<", ">>"],  # noqa: B006
    dynamic_date_format: str = "%Y%m%d",
    dynamic_date_timezone: str = "UTC",
    tests: dict[str, Any] | None = None,
    config_key: str | None = None,
    credentials_secret: str | None = None,
    sql_server_credentials_secret: str | None = None,
    sql_server_config_key: str | None = None,
) -> None:
    """Flow to load data from a SQL Server database to Redshift Spectrum.

    Args:
        query (str): The SQL query to execute on the SQL Server database. Dynamic date
            pattern is supported (SELECT * FROM table WHERE date='<<yesterday>>')
        to_path (str): The path where the extracted data will be stored.
        schema_name (str): The name of the schema in Redshift Spectrum.
        table (str): The name of the table in Redshift Spectrum.
        sep (str, optional): The separator used in the output file. Defaults to ",".
        extension (str, optional): The file extension to use. Defaults to ".parquet".
        if_exists (Literal["overwrite", "append"], optional): Action to take if
            the table already exists. Defaults to "overwrite".
        partition_cols (list[str] | None, optional): Columns to partition the data by.
            Defaults to None.
        index (bool, optional): Whether to include the index in the output file.
            Defaults to False.
        compression (str | None, optional): Compression type to use for the output file.
            Defaults to None.
        dynamic_date_symbols (list[str], optional): Symbols used for dynamic date
            handling. Defaults to ["<<", ">>"].
        dynamic_date_format (str, optional): Format used for dynamic date parsing.
            Defaults to "%Y%m%d".
        dynamic_date_timezone (str, optional): Timezone used for dynamic date
            processing. Defaults to "UTC".
        tests (dict[str, Any], optional): A dictionary with optional list of tests
            to verify the output dataframe. If defined, triggers the `validate`
            function from utils. Defaults to None.
            Available tests:
                - `column_size`: dict{column: size}
                - `column_unique_values`: list[columns]
                - `column_list_to_match`: list[columns]
                - `dataset_row_count`: dict: {'min': number, 'max', number}
                - `column_match_regex`: dict: {column: 'regex'}
                - `column_sum`: dict: {column: {'min': number, 'max': number}}
            Defaults to: None
        config_key (str | None, optional): AWS configuration key. Defaults to None.
        credentials_secret (str | None, optional): Name of the secret storing
            AWS credentials. Defaults to None.
        sql_server_credentials_secret (str | None, optional): Name of the secret storing
            SQL Server credentials. Defaults to None.
        sql_server_config_key (str | None, optional): Key in the configuration for
            SQL Server credentials. Defaults to None.

    Returns:
        None
    """
    ddh = DynamicDateHandler(
        dynamic_date_symbols, dynamic_date_format, dynamic_date_timezone
    )
    query = ddh.process_dates(query)  # type: ignore

    df = sql_server_to_df(
        query=query,
        config_key=sql_server_config_key,
        credentials_secret=sql_server_credentials_secret,
        tests=tests,
    )

    df_to_redshift_spectrum(
        df=df,
        to_path=to_path,
        schema_name=schema_name,
        table=table,
        extension=extension,
        if_exists=if_exists,
        partition_cols=partition_cols,
        index=index,
        compression=compression,
        sep=sep,
        config_key=config_key,
        credentials_secret=credentials_secret,
    )



"""Flows for pulling data from Exchange rates API to Databricks."""

from datetime import datetime
from typing import Literal

from prefect import flow

from viadot.orchestration.prefect.tasks import df_to_databricks, exchange_rates_to_df
from viadot.orchestration.prefect.utils import with_flow_timeout_param


Currency = Literal[
    "USD", "EUR", "GBP", "CHF", "PLN", "DKK", "COP", "CZK", "SEK", "NOK", "ISK"
]


@flow(
    name="extract--exchange-rates-api--databricks",
    description="Extract data from Exchange Rates API and load it into Databricks.",
    retries=1,
    retry_delay_seconds=60,
)
@with_flow_timeout_param()
def exchange_rates_to_databricks(  # noqa: PLR0913
    databricks_table: str,
    databricks_schema: str | None = None,
    if_exists: Literal["replace", "skip", "fail"] = "fail",
    currency: Currency = "USD",
    start_date: str = datetime.today().strftime("%Y-%m-%d"),
    end_date: str = datetime.today().strftime("%Y-%m-%d"),
    symbols: list[str] | None = None,
    exchange_rates_credentials_secret: str | None = None,
    exchange_rates_config_key: str | None = None,
    databricks_credentials_secret: str | None = None,
    databricks_config_key: str | None = None,
) -> None:
    """Download a DataFrame from ExchangeRates API and upload it to Databricks.

    Args:
        databricks_table (str): The name of the target table.
        databricks_schema (str, optional): The name of the target schema.
            Defaults to None.
        if_exists (Literal["replace", "skip", "fail"], optional):
            What to do if the table already exists.
            One of "replace", "skip", and "fail". Defaults to "fail".
        currency (Currency, optional): Base currency to which prices of searched
            currencies are related. Defaults to "USD".
        start_date (str, optional): Initial date for data search.
            Data range is start_date -> end_date,
            supported format 'yyyy-mm-dd'.
            Defaults to datetime.today().strftime("%Y-%m-%d").
        end_date (str, optional): See above.
            Defaults to datetime.today().strftime("%Y-%m-%d").
        symbols (List[str], optional): List of currencies for which
            exchange rates from base currency will be fetched.
            Defaults to
            ["USD","EUR","GBP","CHF","PLN","DKK","COP","CZK","SEK","NOK","ISK"].
            Only ISO codes.
        exchange_rates_credentials_secret (str, optional): The name of the
            Azure Key Vault secret storing the exchange rates credentials.
            Defaults to None.
        exchange_rates_config_key (str, optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
        databricks_credentials_secret (str, optional): The name of the Azure Key Vault
            secret storing relevant credentials. Defaults to None.
        databricks_config_key (str, optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
    """
    df = exchange_rates_to_df(
        currency=currency,
        credentials_secret=exchange_rates_credentials_secret,
        config_key=exchange_rates_config_key,
        start_date=start_date,
        end_date=end_date,
        symbols=symbols,
    )

    df_to_databricks(
        df=df,
        schema=databricks_schema,
        table=databricks_table,
        if_exists=if_exists,
        credentials_secret=databricks_credentials_secret,
        config_key=databricks_config_key,
    )



"""Download data from Mindful API and load it into Azure Data Lake Storage."""

from datetime import date
import time
from typing import Literal

from prefect import flow

from viadot.orchestration.prefect.tasks import df_to_adls, mindful_to_df
from viadot.orchestration.prefect.utils import with_flow_timeout_param


@flow(
    name="Mindful extraction to ADLS",
    description="Extract data from mindful and load it into Azure Data Lake Storage.",
    retries=1,
    retry_delay_seconds=60,
)
@with_flow_timeout_param()
def mindful_to_adls(
    config_key: str | None = None,
    azure_key_vault_secret: str | None = None,
    region: Literal["us1", "us2", "us3", "ca1", "eu1", "au1"] = "eu1",
    endpoint: list[str] | str | None = None,
    date_interval: list[date] | None = None,
    limit: int = 1000,
    adls_config_key: str | None = None,
    adls_azure_key_vault_secret: str | None = None,
    adls_path: str | None = None,
    adls_path_overwrite: bool = False,
) -> None:
    """Flow to download data from Mindful to Azure Data Lake.

    Args:
        credentials (Optional[Dict[str, Any]], optional): Mindful credentials as a
            dictionary. Defaults to None.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        azure_key_vault_secret (Optional[str], optional): The name of the Azure Key
            Vault secret where credentials are stored. Defaults to None.
        region (Literal[us1, us2, us3, ca1, eu1, au1], optional): Survey Dynamix region
            from where to interact with the mindful API. Defaults to "eu1" English
            (United Kingdom).
        endpoint (Optional[Union[List[str], str]], optional): Endpoint name or list of
            them from where to download data. Defaults to None.
        date_interval (Optional[List[date]], optional): Date time range detailing the
            starting date and the ending date. If no range is passed, one day of data
            since this moment will be retrieved. Defaults to None.
        limit (int, optional): The number of matching interactions to return.
            Defaults to 1000.
        adls_credentials (Optional[Dict[str, Any]], optional): The credentials as a
            dictionary. Defaults to None.
        adls_config_key (Optional[str], optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
        adls_azure_key_vault_secret (Optional[str], optional): The name of the Azure Key
            Vault secret containing a dictionary with ACCOUNT_NAME and Service Principal
            credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
            Defaults to None.
        adls_path (Optional[str], optional): Azure Data Lake destination file path.
            Defaults to None.
        adls_path_overwrite (bool, optional): Whether to overwrite the file in ADLS.
            Defaults to True.

    Examples:
        mindful_to_adls(
            config_key=config_key,
            endpoint=endpoint,
            date_interval=date_interval,
            adls_path=adls_path,
            adls_config_key=adls_config_key,
            adls_azure_key_vault_secret=adls_azure_key_vault_secret,
            adls_path_overwrite=True,
        )
    """
    if isinstance(endpoint, str):
        endpoint = [endpoint]

    endpoints = endpoint

    for endpoint in endpoints:
        data_frame = mindful_to_df(
            config_key=config_key,
            azure_key_vault_secret=azure_key_vault_secret,
            region=region,
            endpoint=endpoint,
            date_interval=date_interval,
            limit=limit,
        )

        # ???
        time.sleep(0.5)

        df_to_adls(
            df=data_frame,
            path=adls_path.rstrip("/") + "/" + f"{endpoint}.csv",
            credentials_secret=adls_azure_key_vault_secret,
            config_key=adls_config_key,
            overwrite=adls_path_overwrite,
        )



from unittest.mock import patch

import pandas as pd
import pytest

from viadot.exceptions import ValidationError
from viadot.sources import TM1
from viadot.sources.tm1 import TM1Credentials


data = {"a": [420, 380, 390], "b": [50, 40, 45]}
test_cube = "test_cube"
test_view = "test_view"
test_dimension = "test_dim"
test_hierarchy = "test_hier"


@pytest.fixture
def tm1_credentials():
    return TM1Credentials(
        username="test_user",
        password="test_password",  # pragma: allowlist secret # noqa: S106
        address="localhost",
        port="12356",
    )


@pytest.fixture
def tm1(tm1_credentials: TM1Credentials):
    return TM1(
        credentials=tm1_credentials.dict(),
        verify=False,
    )


def test_tm1_initialization(tm1):
    """Test that the TM1 object is initialized with the correct credentials."""
    assert tm1.credentials.get("address") == "localhost"
    assert tm1.credentials.get("username") == "test_user"
    assert (
        tm1.credentials.get("password").get_secret_value()
        == "test_password"  # pragma: allowlist secret
    )


def test_get_cubes_names(tm1):
    with patch("viadot.sources.tm1.TM1Service") as mock:
        instance = mock.return_value
        instance.cubes.get_all_names.return_value = ["msg", "cuve"]
        result = tm1.get_cubes_names()
        assert result == ["msg", "cuve"]


def test_get_views_names(tm1):
    with patch("viadot.sources.tm1.TM1Service") as mock:
        instance = mock.return_value
        instance.views.get_all_names.return_value = ["view1", "view1"]
        result = tm1.get_views_names(cube=test_cube)
        assert result == ["view1", "view1"]


def test_get_dimensions_names(tm1):
    with patch("viadot.sources.tm1.TM1Service") as mock:
        instance = mock.return_value
        instance.dimensions.get_all_names.return_value = ["dim1", "dim2"]
        result = tm1.get_dimensions_names()
        assert result == ["dim1", "dim2"]


def test_get_available_elements(tm1):
    with patch("viadot.sources.tm1.TM1Service") as mock:
        instance = mock.return_value
        instance.elements.get_element_names.return_value = ["el1", "el2"]
        result = tm1.get_available_elements(
            dimension=test_dimension, hierarchy=test_hierarchy
        )
        assert result == ["el1", "el2"]


def test_to_df(tm1):
    with patch("viadot.sources.tm1.TM1Service") as mock:
        instance = mock.return_value
        instance.cubes.cells.execute_view_dataframe.return_value = pd.DataFrame(data)
        result = tm1.to_df(cube=test_cube, view=test_view)
        assert isinstance(result, pd.DataFrame)
        assert not result.empty
        assert len(result.columns) == 4


def test_to_df_fail(tm1, mocker):
    mocker.patch("viadot.sources.TM1.get_connection", return_value=True)
    with pytest.raises(ValidationError) as excinfo:
        tm1.to_df()
    assert str(excinfo.value) == "MDX query or cube and view are required."



"""Utility tasks."""

import json
from pathlib import Path
from typing import Any, Literal

import pandas as pd
from prefect import task
from prefect.logging import get_run_logger
from visions.functional import infer_type
from visions.typesets.complete_set import CompleteSet


@task
def dtypes_to_json_task(dtypes_dict: dict[str, Any], local_json_path: str) -> None:
    """Creates json file from a dictionary.

    Args:
        dtypes_dict (dict): Dictionary containing data types.
        local_json_path (str): Path to local json file.
    """
    with Path(local_json_path).open("w") as file_path:
        json.dump(dtypes_dict, file_path)


@task
def chunk_df(df: pd.DataFrame, size: int = 10_000) -> list[pd.DataFrame]:
    """Chunks a data frame into multiple smaller data frames of a specified size.

    Args:
        df (pd.DataFrame): Input pandas DataFrame.
        size (int, optional): Size of a chunk. Defaults to 10000.
    """
    n_rows = df.shape[0]
    return [df[i : i + size] for i in range(0, n_rows, size)]


@task
def df_get_data_types_task(df: pd.DataFrame) -> dict:
    """Returns dictionary containing datatypes of pandas DataFrame columns.

    Args:
        df (pd.DataFrame): Input pandas DataFrame.
    """
    typeset = CompleteSet()
    dtypes = infer_type(df, typeset)
    return {k: str(v) for k, v in dtypes.items()}


@task
def get_sql_dtypes_from_df(df: pd.DataFrame) -> dict:
    """Obtain SQL data types from a pandas DataFrame."""
    typeset = CompleteSet()
    dtypes = infer_type(df.head(10000), typeset)
    dtypes_dict = {k: str(v) for k, v in dtypes.items()}
    dict_mapping = {
        "Float": "REAL",
        "Image": None,
        "Categorical": "VARCHAR(500)",
        "Time": "TIME",
        "Boolean": "VARCHAR(5)",  # Bool is True/False, Microsoft expects 0/1
        "DateTime": "DATETIMEOFFSET",  # DATETIMEOFFSET is timezone-aware dtype in TSQL
        "Object": "VARCHAR(500)",
        "EmailAddress": "VARCHAR(50)",
        "File": None,
        "Geometry": "GEOMETRY",
        "Ordinal": "INT",
        "Integer": "INT",
        "Generic": "VARCHAR(500)",
        "UUID": "VARCHAR(50)",  # Microsoft uses a custom UUID format so we can't use it
        "Complex": None,
        "Date": "DATE",
        "String": "VARCHAR(500)",
        "IPAddress": "VARCHAR(39)",
        "Path": "VARCHAR(255)",
        "TimeDelta": "VARCHAR(20)",  # datetime.datetime.timedelta; eg.'1 days 11:00:00'
        "URL": "VARCHAR(255)",
        "Count": "INT",
    }
    dict_dtypes_mapped = {}
    for k in dtypes_dict:
        dict_dtypes_mapped[k] = dict_mapping[dtypes_dict[k]]

    # This is required as pandas cannot handle mixed dtypes in Object columns
    return {
        k: ("String" if v == "Object" else str(v))
        for k, v in dict_dtypes_mapped.items()
    }


@task
def df_map_mixed_dtypes_for_parquet(
    df: pd.DataFrame, dtypes_dict: dict
) -> pd.DataFrame:
    """Handle mixed dtypes in DataFrame columns.

    Mapping 'object' visions dtype to 'string' dtype to allow Pandas to_parquet

    Args:
        dict_dtypes_mapped (dict): Data types dictionary, inferred by Visions.
        df (pd.DataFrame): input DataFrame.

    Returns:
        df_mapped (pd.DataFrame): Pandas DataFrame with mapped Data Types to workaround
        Pandas to_parquet bug connected with mixed dtypes in object:.
    """
    df_mapped = df.copy()
    for col, dtype in dtypes_dict.items():
        if dtype == "Object":
            df_mapped[col] = df_mapped[col].astype("string")
    return df_mapped


@task
def update_dtypes_dict(dtypes_dict: dict) -> dict:
    """Task to update dtypes_dictionary that will be stored in the schema.

    It's required due to workaround Pandas to_parquet bug connected with mixed dtypes in
    object.

    Args:
        dtypes_dict (dict): Data types dictionary inferred by Visions.

    Returns:
        dtypes_dict_updated (dict): Data types dictionary updated to follow Pandas
        requirements in to_parquet functionality.
    """
    return {k: ("String" if v == "Object" else str(v)) for k, v in dtypes_dict.items()}


@task
def df_to_csv(
    df: pd.DataFrame,
    path: str,
    sep: str = "\t",
    if_exists: Literal["append", "replace", "skip"] = "replace",
    **kwargs,
) -> None:
    r"""Write data from a pandas DataFrame to a CSV file.

    Args:
        df (pd.DataFrame): Input pandas DataFrame.
        path (str): Path to output csv file.
        sep (str, optional): The separator to use in the CSV. Defaults to "\t".
        if_exists (Literal["append", "replace", "skip"], optional): What to do if the
            table exists. Defaults to "replace".
    """
    logger = get_run_logger()

    if Path(path).exists():
        if if_exists == "append":
            existing_df = pd.read_csv(path, sep=sep)
            out_df = pd.concat([existing_df, df])
        elif if_exists == "replace":
            out_df = df
        elif if_exists == "skip":
            logger.info("Skipped.")
            return
    else:
        out_df = df

    # Create directories if they don't exist.
    Path(path).parent.mkdir(parents=True, exist_ok=True)

    out_df.to_csv(path, index=False, sep=sep, **kwargs)


@task
def df_to_parquet(
    df: pd.DataFrame,
    path: str,
    if_exists: Literal["append", "replace", "skip"] = "replace",
    **kwargs,
) -> None:
    """Task to create parquet file based on pandas DataFrame.

    Args:
        df (pd.DataFrame): Input pandas DataFrame.
        path (str): Path to output parquet file.
        if_exists (Literal["append", "replace", "skip"], optional): What to do if the
            table exists. Defaults to "replace".
    """
    logger = get_run_logger()

    if Path(path).exists():
        if if_exists == "append":
            existing_df = pd.read_parquet(path)
            out_df = pd.concat([existing_df, df])
        elif if_exists == "replace":
            out_df = df
        elif if_exists == "skip":
            logger.info("Skipped.")
            return
    else:
        out_df = df

    # Create directories if they don't exist.
    Path(path).parent.mkdir(parents=True, exist_ok=True)

    out_df.to_parquet(path, index=False, **kwargs)


@task
def union_dfs_task(dfs: list[pd.DataFrame]) -> pd.DataFrame:
    """Create one DataFrame from a list of pandas DataFrames.

    Args:
        dfs (List[pd.DataFrame]): List of pandas data frames to concat. In case of
            different size of DataFrames NaN values can appear.
    """
    return pd.concat(dfs, ignore_index=True)



"""Build specified dbt model(s) and upload the generated metadata to Luma."""

from datetime import datetime, timezone
from pathlib import Path
import re
import shutil
from typing import Literal

from prefect import flow, task
from prefect.logging import get_run_logger
from prefect.states import Failed, State

from viadot.orchestration.prefect.tasks import (
    SmtpConfig,
    clone_repo,
    dbt_task,
    dbt_test_failure_notifier,
    luma_ingest_task,
    s3_upload_file,
)
from viadot.orchestration.prefect.utils import get_credentials, with_flow_timeout_param


@task(cache_policy=None)
def remove_dbt_repo_dir(dbt_repo_dir_name: str) -> None:
    """Remove the repo directory.

    Args:
        dbt_repo_dir_name (str): The name of the dbt repo directory.
    """
    shutil.rmtree(dbt_repo_dir_name, ignore_errors=True)


@flow(
    name="Transform and Catalog",
    description="Build specified dbt model(s) and upload generated metadata to Luma.",
)
@with_flow_timeout_param()
def transform_and_catalog(  # noqa: PLR0913, PLR0915
    dbt_repo_url: str | None = None,
    dbt_repo_url_secret: str | None = None,
    dbt_project_path: str = "dbt",
    dbt_repo_branch: str | None = None,
    dbt_repo_token_secret: str | None = None,
    dbt_selects: dict[str, str] | None = None,
    dbt_target: str | None = None,
    dbt_target_dir_path: str | None = None,
    luma_url: str = "http://localhost:8000",
    luma_follow: bool = False,
    metadata_kind: Literal["model", "model_run"] = "model_run",
    run_results_storage_path: str | None = None,
    run_results_storage_config_key: str | None = None,
    run_results_storage_credentials_secret: str | None = None,
    fail_flow_only_on_build_failure: bool = False,
    additional_recipients: list[str] | None = None,
    recipients: list[str] | None = None,
    smtp_config: SmtpConfig | None = None,
    enable_notifications: bool = True,
    gh_action_actor: str | None = None,
) -> State | None:
    """Build specified dbt model(s) and upload the generated metadata to Luma.

    Supports ingesting both model and model run metadata (controlled by the
    `metadata_kind` parameter).

    Note that metadata is still ingested even if the preceding `dbt test` task fails.
    This is done in order to capture test failure metadata in the data catalog.

    Args:
        dbt_repo_url (str, optional): The URL for cloning the dbt repo with relevant
            dbt project. Defaults to None.
        dbt_repo_url_secret (str, optional): Alternatively to above, the secret
            containing `dbt_repo_url`. Defaults to None.
        dbt_project_path (str): Path to the dbt project directory, relative to the
            dbt repository's root. For example, "dbt/my_dbt_project". Defaults to "dbt".
        dbt_repo_branch (str, optional): The branch of the dbt repo to use. Defaults to
            None (default repo branch).
        dbt_repo_token_secret (str, optional): The secret containing the personal access
            token used to clone the dbt repository, in case it's private. Not required
            if token is already included in `dbt_repo_url` (which is NOT recommended).
            Defaults to None.
        dbt_selects (dict, optional): Valid
            [dbt node selection](https://docs.getdbt.com/reference/node-selection/syntax)
            expressions. Valid keys are `run`, `test`,`build`, and `source_freshness`.
                The test select expression is taken from run's, as long as run select is
                provided. Defaults to None.
        dbt_target (str): The dbt target to use. If not specified, the default dbt
            target (as specified in `profiles.yaml`) will be used. Defaults to None.
        dbt_target_dir_path (str): The path to your dbt project's target
            directory, which contains dbt artifact JSON files, relative
            to dbt project's root directory. By default,
            `<repo_name>/<dbt_project_path>/target`, since "target" is the default
            name of the directory generated by dbt.
        luma_url (str, optional): The URL of the Luma instance to ingest into.
            Defaults to "http://localhost:8000".
        luma_follow (bool, optional): Whether to follow the ingestion process until it's
            completed (by default, ingestion request is sent without awaiting for the
            response). By default, `False`.
        metadata_kind (Literal["model", "model_run"], optional): The kind of metadata
            to ingest. Defaults to "model_run".
        run_results_storage_path (str, optional): The directory to upload the
            `run_results.json` file to. Note that a timestamp will be appended to the
            end of the file. Currently, only S3 is supported. Defaults to None.
        run_results_storage_config_key (str, optional): The key in the viadot config
            holding AWS credentials. Defaults to None.
        run_results_storage_credentials_secret (str, optional): The name of the secret
            block in Prefect holding AWS credentials. Defaults to None.
        fail_flow_only_on_build_failure (bool): Determines the flow's failure behavior
            based on dbt build outcomes.
            When False (default):
                - The flow will fail on any dbt build failure (including test failures)
            When True:
                - The flow will only fail if model building fails
                - Test failures alone won't cause the flow failure
        recipients (list[str] | None, optional): Primary recipient list. If provided,
            it takes precedence over the extracted owners email addresses from dbt
            metadata. Defaults to None.
        additional_recipients (list[str] | None, optional): Extra email addresses
            to be appended to the final recipient list regardless of other settings.
            Defaults to None.
        enable_notifications (bool): Whether to send email notifications for failed
            dbt tests. If True, notifications are sent to Technical Owners defined
            in dbt's meta config, with `notification_recipients` used as
            a fallback when no owner is found.
            If False, no notifications are sent regardless of
            `notification_recipients`. Defaults to True.
        gh_action_actor (str, optional): GitHub Actions actor that triggered the
            workflow. Defaults to None.

    Returns:
        list[str]: Lines from stdout of the `upload_metadata` task as a list.

    Examples:
        # Build staging models.

        ```python
        import os
        from prefect_viadot.flows import transform_and_catalog

        my_dbt_repo_url = "https://github.com/dbt-labs/jaffle_shop"
        my_luma_url = "http://localhost:8000"

        transform_and_catalog(
            dbt_repo_url=my_dbt_repo_url
            dbt_selects={"run": "staging"}
            luma_url=my_luma_url,
            run_results_storage_path="s3://my-bucket/dbt/run_results",
            run_results_storage_credentials_secret="my-aws-credentials-block",
        )
        ```

        Some common `dbt_selects` patterns:
        - runs a specific model and all its downstream dependencies:
            `dbt_select={"run": "my_model+"}`
        - runs all models in a directory:
            `dbt_select={"run: "models/staging"}`
        - runs a specific model in a folder:
            `dbt_select={"run": "marts.domain.some_model"}`
        - runs tests for a specific model:
            `dbt_select={"test": "my_model"}`
        - build a specific model:
            `dbt_select={"build": "my_model"}`
        - build all models in a folder:
            `dbt_select={"build": "models.intermediate"}`
    """
    logger = get_run_logger()
    if gh_action_actor:
        logger.info(f"Triggered by GitHub Actions actor: {gh_action_actor}")

    # Clone the dbt project.
    dbt_repo_url = dbt_repo_url or get_credentials(dbt_repo_url_secret)
    clone_repo(
        url=dbt_repo_url,
        checkout_branch=dbt_repo_branch,
        token_secret=dbt_repo_token_secret,
    )

    # Prepare the environment.
    dbt_repo_name = dbt_repo_url.split("/")[-1].replace(".git", "")
    dbt_project_path_full = Path(dbt_repo_name) / dbt_project_path
    dbt_pull_deps_task = dbt_task.with_options(
        name="dbt_deps", retries=3, retry_delay_seconds=60
    )
    pull_dbt_deps = dbt_pull_deps_task.submit(
        project_path=dbt_project_path_full,
        command="deps",
    )
    pull_dbt_deps.result()

    # Run dbt commands.
    dbt_target_option = f"-t {dbt_target}" if dbt_target is not None else ""

    if metadata_kind == "model_run":
        # Produce `run-results.json` artifact for Luma ingestion.
        if dbt_selects:
            build_select = dbt_selects.get("build")
            run_select = dbt_selects.get("run")
            test_select = dbt_selects.get("test", run_select)

            build_select_safe = f"-s {build_select}" if build_select is not None else ""
            run_select_safe = f"-s {run_select}" if run_select is not None else ""
            test_select_safe = f"-s {test_select}" if test_select is not None else ""
        else:
            run_select_safe = ""
            test_select_safe = ""
        if build_select:
            # If build task is used, run and test tasks are not needed.
            # Build task executes run and tests commands internally.
            build_task = dbt_task.with_options(name="dbt_build")
            raise_on_failure = not fail_flow_only_on_build_failure

            build = build_task.submit(
                project_path=dbt_project_path_full,
                command=f"build {build_select_safe} {dbt_target_option}",
                raise_on_failure=raise_on_failure,
                return_all=True,
            )
            build.result()
        else:
            run_task = dbt_task.with_options(name="dbt_run")
            run = run_task.submit(
                project_path=dbt_project_path_full,
                command=f"run {run_select_safe} {dbt_target_option}",
            )
            run.result()

            test_task = dbt_task.with_options(name="dbt_test")
            test = test_task.submit(
                project_path=dbt_project_path_full,
                command=f"test {test_select_safe} {dbt_target_option}",
                raise_on_failure=False,
            )
            test.result()
    else:
        # Produce `catalog.json` and `manifest.json` artifacts for Luma ingestion.
        docs_generate_task = dbt_task.with_options(name="dbt_docs_generate")
        docs = docs_generate_task.submit(
            project_path=dbt_project_path_full,
            command="docs generate",
        )
        docs.result()

    # Upload metadata to Luma.
    if dbt_target_dir_path is None:
        dbt_target_dir_path = dbt_project_path_full / "target"

    upload_metadata = luma_ingest_task.submit(
        metadata_kind=metadata_kind,
        metadata_dir_path=dbt_target_dir_path,
        luma_url=luma_url,
        follow=luma_follow,
    )
    upload_metadata.result()

    file_name = "run_results.json"
    run_results_file_path = str(dbt_target_dir_path / file_name)  # type: ignore
    if run_results_storage_path:
        # Set the file path to include date info.
        now = datetime.now(timezone.utc)
        run_results_storage_path = run_results_storage_path.rstrip("/") + "/"

        # Add partitioning.
        date_str = now.strftime("%Y%m%d")
        run_results_storage_path += date_str + "/"

        # Add timestamp suffix, eg. run_results_1737556947.934292.json.
        timestamp = now.timestamp()
        run_results_storage_path += (
            Path(file_name).stem + "_" + str(timestamp) + ".json"
        )
        logger.info(f"Uploading run results to {run_results_storage_path}")
        # Upload the file to s3.

        s3_upload_file(
            from_path=run_results_file_path,
            to_path=run_results_storage_path,
            config_key=run_results_storage_config_key,
            credentials_secret=run_results_storage_credentials_secret,
        )
    if enable_notifications:
        dbt_test_failure_notifier(
            results_file_path=run_results_file_path,
            manifest_file_path=str(dbt_target_dir_path / "manifest.json"),
            recipients=recipients,
            additional_recipients=additional_recipients,
            smtp_config=smtp_config,
        )

    remove_dbt_repo_dir(
        dbt_repo_dir_name=dbt_repo_name,
    )

    if fail_flow_only_on_build_failure and build_select:
        build_result = build.result()
        model_error_pattern = re.compile(r"ERROR creating", re.IGNORECASE)
        if any(model_error_pattern.search(line) for line in build_result):
            return Failed(message="One or more models failed to build.")

    return None



"""Module for sending email notifications about failed dbt tests."""

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import json
from pathlib import Path
import re
import smtplib

import pandas as pd
from prefect import task
from prefect.blocks.system import Secret
from prefect.logging import get_run_logger
from pydantic import BaseModel


class SmtpConfig(BaseModel):
    host: str = "smtp.gmail.com"
    port: int = 587
    sender: str
    password: str


def find_schema(compiled_code: str | None) -> str | None:
    """Extract a test schema name from a string from a pandas Series.

        This function searches for a pattern in each element of the pandas Series
        from the "compiled_code" column.It handles strings with extra characters
        like new line characters etc.

    Args:
        compiled_code (str | None):  SQL string (or None) to extract
        the schema name from

    Returns:
        str: A string containing the extracted patterns or None if not found.
    """
    if compiled_code is None:
        return None
    pattern = r'[\"a-zA-Z0-9_\"]\.\"([^"]+)\"\.'
    match = re.search(pattern, compiled_code)
    return match.group(1) if match else None


def find_model(compiled_code: str | None) -> str | None:
    """Extract a test model name from a string from a pandas Series.

    This function searches for a pattern in each element of the pandas Series
    from the "compiled_code" column.It handles strings with extra characters
    like new line characters etc.

    Args:
        compiled_code (str | None): SQL string (or None) to extract
        the table name from

    Returns:
        str: A string containing the extracted patterns or None if not found.
    """
    if compiled_code is None:
        return None
    pattern = r"[\"a-zA-Z0-9_\"]\.+[\"a-zA-Z0-9_\"]+\".+\"([a-zA-Z0-9_]+?)\""
    match = re.search(pattern, compiled_code)
    return match.group(1) if match else None


def find_column(unique_id: str | None) -> str | None:
    """Extract a test column name from a string from a pandas Series.

    This function searches for a pattern in each element of the pandas Series
    from the "unique_id" column.

    Args:
        unique_id (str | None): unique_id to string (or None) to extract
        the column name from

    Returns:
        str: A string containing the extracted patterns or None if not found.
    """
    if unique_id is None:
        return None
    pattern = r".*?__([a-zA-Z0-9_]+?)(?=_?\.[a-zA-Z0-9]+$|__)"
    match = re.search(pattern, unique_id)
    return match.group(1) if match else None


def find_test(value: str | None, test_types: tuple[str, ...]) -> str | None:
    """Extract a specific test name from a string in pandas Series.

    This function searches for allowed test types in each element
        of the pandas Series(df column).

    Args:
        value (str | None): The string (or None) to search within.
        test_types: (tuple[str, ...]): A list of test types that can be discovered.

    Returns:
        str | None: The matched test type, or None if no match is found.
    """
    if value is None:
        return None
    for test_type in test_types:
        if test_type in value:
            return test_type
    return None


def extract_model_ownership(manifest_file_path: str) -> pd.DataFrame:
    """Extract model ownership information from a dbt manifest.json file."""
    with Path(manifest_file_path).open() as file:
        manifest = json.load(file)

    result = []
    for _, node in manifest["nodes"].items():
        if node.get("resource_type") != "model":
            continue

        owners = node.get("config", {}).get("meta", {}).get("owners", [])
        technical_owner = next(
            (o["email"] for o in owners if o.get("type") == "Technical Owner"), None
        )

        result.append(
            {
                "model": node["name"],
                "schema": node["schema"],
                "database": node["database"],
                "technical_owner": technical_owner,
            }
        )

    df = pd.DataFrame(result)
    return df.drop_duplicates(subset=["model", "schema", "technical_owner"])


def get_owner_emails(
    failed_model: str, failed_schema: str, df: pd.DataFrame
) -> list[str]:
    """Get owner emails based on failed model and schema."""
    model_emails = (
        df.loc[df["model"] == failed_model, "technical_owner"]
        .dropna()
        .unique()
        .tolist()
    )
    if model_emails:
        return model_emails
    return (
        df.loc[df["schema"] == failed_schema, "technical_owner"]
        .dropna()
        .unique()
        .tolist()
    )


def enrich_with_owners(failed_tests: list, owners_df: pd.DataFrame) -> list:
    """Separate function responsible for enriching data with owners."""
    for test in failed_tests:
        test["owners"] = get_owner_emails(test["model"], test["schema"], owners_df)
    return failed_tests


def extract_failed_tests(
    run_results_file_path: str, test_types: tuple[str, ...]
) -> list:
    """Extract failed tests from `run_results.json`."""
    with Path(run_results_file_path).open() as file:
        data = json.load(file)

    df = pd.json_normalize(
        data,
        meta=[["metadata", "generated_at"]],
        record_path="results",
    )
    if df.empty:
        return []
    test_identifier = "^test\\."
    contains_test = df["unique_id"].str.match(test_identifier)
    if not contains_test.any():
        return []

    if not contains_test.all():
        df = df[contains_test].copy()

    df_failed = df[df["status"].isin(["error", "fail"])].copy()
    df_failed["schema"] = df_failed["compiled_code"].apply(find_schema)
    df_failed["model"] = df_failed["compiled_code"].apply(find_model)
    df_failed["column"] = df_failed["unique_id"].apply(find_column)
    df_failed["test_type"] = df_failed["unique_id"].apply(
        lambda x: find_test(x, test_types)
    )

    if df_failed.empty:
        return []

    return (
        df_failed[
            [
                "status",
                "model",
                "column",
                "test_type",
                "failures",
                "message",
                "schema",
                "unique_id",
            ]
        ]
        .rename(columns={"unique_id": "test_code"})
        .to_dict(orient="records")
    )


def extract_subject_from_message(message: str) -> str:
    """Extract model name from dbt error message as email subject."""
    match = re.search(r"test\s+(\S+)\s+\(", message)
    if match:
        return f"DBT Test Alert: {match.group(1)}"
    return "DBT Test Alert"


def dataframe_to_email_html(df: pd.DataFrame) -> str:
    """Convert DataFrame to a styled single-row HTML table for email."""
    styles = {
        "table": "border-collapse: collapse; font-family: Arial, sans-serif; font-size: 13px; width: 100%; border: 1px solid #000000;",
        "th": "background-color: #2c3e50; color: #ffffff; padding: 8px 12px; text-align: left; white-space: nowrap; border: 1px solid #000000;",
        "td": "padding: 8px 12px; border: 1px solid #000000; vertical-align: top;",
        "tr_even": "background-color: #f8f9fa;",
    }

    headers = "".join(f'<th style="{styles["th"]}">{col}</th>' for col in df.columns)

    rows = ""
    for i, (_, row) in enumerate(df.iterrows()):
        bg = f' style="{styles["tr_even"]}"' if i % 2 == 0 else ""
        cells = "".join(
            f'<td style="{styles["td"]}">{val if pd.notna(val) else "N/A"}</td>'
            for val in row
        )
        rows += f"<tr{bg}>{cells}</tr>"

    return f'<table style="{styles["table"]}"><thead><tr>{headers}</tr></thead><tbody>{rows}</tbody></table>'


def send_test_failure_notification(
    failed_test: pd.DataFrame,
    sender: str,
    server: smtplib.SMTP,
    recipients: list[str] | None = None,
    additional_recipients: list[str] | None = None,
) -> None:
    """Send an email notification for a failed DBT test.

    Args:
        failed_test (pd.DataFrame): DataFrame containing details of the failed test.
        sender (str): The email address to appear in the "From" field.
        server (smtplib.SMTP): A pre-configured and authenticated SMTP server
            instance used to send the email.
        recipients (list[str] | None, optional): Primary recipient list. If provided,
            it takes precedence over the 'owners' listed in the DataFrame.
            Defaults to None.
        additional_recipients (list[str] | None, optional): Extra email addresses
            to be appended to the final recipient list regardless of other settings.
            Defaults to None.
    """
    columns_to_skip = {"owners"}
    schema_name = (
        failed_test["schema"].iloc[0]
        if not pd.isna(failed_test["schema"].iloc[0])
        else "N/A"
    )
    model_name = (
        failed_test["model"].iloc[0]
        if not pd.isna(failed_test["model"].iloc[0])
        else "N/A"
    )
    column_name = (
        failed_test["column"].iloc[0]
        if not pd.isna(failed_test["column"].iloc[0])
        else "N/A"
    )
    owners = failed_test["owners"].explode().dropna()
    yaml_owners = owners[owners.str.strip() != ""].unique().tolist()
    # Prevent TypeError when recipients is None
    additional_recipients = additional_recipients or []
    base_recipients = recipients or yaml_owners
    all_recipients = list(set(base_recipients + additional_recipients))
    if not all_recipients:
        return
    recipients_str = ", ".join(all_recipients)

    if any(v == "N/A" for v in [schema_name, column_name, model_name]):
        subject = extract_subject_from_message(failed_test["message"].iloc[0])
    else:
        subject = f"DBT Test Alert: {schema_name} - {model_name}"

    table_html = dataframe_to_email_html(failed_test.drop(columns=columns_to_skip))

    body_html = f"""
    <html>
    <body>
        <h2>DBT Test Failure Notification</h2>
        <p><strong>Recipients:</strong> {recipients_str}</p>
        {table_html}
    </body>
    </html>
    """

    msg = MIMEMultipart("mixed")
    msg["From"] = sender
    msg["To"] = recipients_str
    msg["Subject"] = subject
    msg.attach(MIMEText(body_html, "html"))
    server.sendmail(sender, all_recipients, msg.as_string())


@task(
    name="dbt-test-failure-notifier",
    retries=3,
    retry_delay_seconds=10,
    cache_policy=None,
)
def dbt_test_failure_notifier(
    results_file_path: str,
    manifest_file_path: str,
    additional_recipients: list[str] | None,
    recipients: list[str] | None,
    smtp_config: SmtpConfig | None = None,
    test_types: tuple[str, ...] = (
        "not_null",
        "unique",
        "accepted_values",
        "relationships",
    ),
) -> None:
    """Prefect task to send email notifications for failed DBT tests."""
    logger = get_run_logger()
    if not Path(results_file_path).exists():
        logger.warning(f"File {results_file_path} does not exist.")
        return

    failed_tests = extract_failed_tests(results_file_path, test_types)  # type: ignore
    if not failed_tests:
        logger.info("No failed tests — skipping notifications.")
        return
    owners_df = extract_model_ownership(manifest_file_path)
    failed_tests = enrich_with_owners(failed_tests, owners_df)

    df_failed_tests = pd.DataFrame(failed_tests)
    dfs_list = [group for _, group in df_failed_tests.groupby("model")]

    if smtp_config is None:
        smtp_config = SmtpConfig(
            sender=Secret.load("smtp-sender").get(),
            password=Secret.load("smtp-password").get(),
        )

    with smtplib.SMTP(smtp_config.host, smtp_config.port) as server:
        server.starttls()
        server.login(smtp_config.sender, smtp_config.password)
        sent = 0
        for single_model_tests in dfs_list:
            try:
                send_test_failure_notification(
                    single_model_tests,
                    smtp_config.sender,
                    server,
                    additional_recipients,
                    recipients,
                )
                sent += 1
            except smtplib.SMTPException:
                logger.exception(
                    f"Failed to send for {single_model_tests['model'].iloc[0]}"
                )
            except Exception:
                logger.exception(
                    f"Unexpected error for {single_model_tests['model'].iloc[0]}"
                )
        logger.info(f"Sent {sent}/{len(failed_tests)} failure notification(s).")



from prefect import flow

from viadot.orchestration.prefect.tasks import dbt_task


def test_dbt_task():
    @flow
    def test_flow():
        return dbt_task()

    result = test_flow()
    assert result == "Hello, prefect-viadot!"



"""Task for uploading pandas DataFrame to MinIO."""

import contextlib
from typing import Literal

import pandas as pd
from prefect import task
from prefect.logging import get_run_logger


with contextlib.suppress(ImportError):
    from viadot.sources import MinIO

from viadot.config import get_source_credentials
from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60)
def df_to_minio(
    df: pd.DataFrame,
    path: str,
    credentials_secret: str | None = None,
    config_key: str | None = None,
    basename_template: str | None = None,
    if_exists: Literal["error", "delete_matching", "overwrite_or_ignore"] = "error",
) -> None:
    """Task for uploading the contents of a pandas DataFrame to MinIO.

    Args:
        df (pd.DataFrame): Pandas dataframe to be uploaded.
        path (str): Path to the MinIO file/folder.
        credentials_secret (str, optional): The name of the secret storing
            the credentials. Defaults to None. More info on:
            https://docs.prefect.io/concepts/blocks/
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        basename_template (str, optional): A template string used to generate
            basenames of written data files. The token '{i}' will be replaced with
            an automatically incremented integer. Defaults to None.
        if_exists (Literal["error", "delete_matching", "overwrite_or_ignore"],
            optional). What to do if the dataset already exists. Defaults to "error".
    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    logger = get_run_logger()

    credentials = get_source_credentials(config_key) or get_credentials(
        credentials_secret
    )
    minio = MinIO(credentials=credentials)

    minio.from_df(
        df=df, path=path, if_exists=if_exists, basename_template=basename_template
    )

    logger.info("Data has been uploaded successfully.")



from pathlib import Path

import pytest

from viadot.orchestration.prefect.flows import duckdb_to_sql_server
from viadot.orchestration.prefect.tasks import sql_server_query
from viadot.sources import DuckDB, SQLServer


TABLE = "test_table"
SCHEMA = "sandbox"
DATABASE_PATH = "test_db_123.duckdb"
DUCKDB_CREDS = {"database": DATABASE_PATH, "read_only": False}


@pytest.fixture
def sql_server():
    # Initialize the SQLServer instance with the test credentials.
    return SQLServer(config_key="sql_server")


@pytest.fixture
def duckdb():
    # Initialize the SQLServer instance with the test credentials.
    duckdb = DuckDB(credentials=DUCKDB_CREDS)
    duckdb.run_query(
        query="""
CREATE SCHEMA sandbox;
CREATE or replace TABLE sandbox.numbers AS
SELECT 42 AS i, 84 AS j;
"""
    )
    yield duckdb
    Path(DATABASE_PATH).unlink()


def test_duckdb_to_sql_server(duckdb, sql_server):  # noqa: ARG001
    duckdb_to_sql_server(
        query="select * from sandbox.numbers",
        local_path="testing.csv",
        db_table=TABLE,
        db_schema=SCHEMA,
        duckdb_credentials=DUCKDB_CREDS,
        sql_server_credentials_secret="sql-server",  # noqa: S106
    )
    assert sql_server.exists(table=TABLE, schema=SCHEMA)

    sql_server_query(
        query=f"""DROP TABLE {SCHEMA}.{TABLE}""",
        credentials_secret="sql-server",  # noqa: S106
    )
